/*
** dump.c - utility to display or modify contents of memory windows on an
**          Alpha Data Gen 3 reconfigurable computing device.
**
** (C) Copyright 2016 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   define _stscanf_s(buffer, ...) _stscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   define _tcscpy_s(x, y, z) tcscpy(x, z)
#   define _tcsnset_s(buffer, max, c, n) _tcsnset(buffer, c, n)
#endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tscanf_s scanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcslen(x) strlen(x)
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _totlower(c) tolower(c)
# define _gettc getc

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <assert.h>
#include <ctype.h>
#include <limits.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <num_token.h>
#include <platform_endian.h>

/* Exit codes */
#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_UNRECOGNIZED_COMMAND    (2) /* Invalid command passed as 1st positional argument */
#define EXIT_ALLOCATION_FAILURE      (3) /* A memory allocation failed */
#define EXIT_DEVICE_OPEN_ERROR       (4) /* Failed to open ADMXRC3 device */
#define EXIT_MAPWINDOW_ERROR         (5) /* Failed to map window into process' virtual address space */

#ifdef _WIN32
# define PROGRAM_NAME _T("dump.exe")
#else
# define PROGRAM_NAME _T("dump")
#endif

#if !defined(SIZE_MAX)
/* Hack for when SIZE_MAX missing; relies on two's complement representation */
# define SIZE_MAX ((size_t)(-1))
#endif

#if defined(_UNICODE)
# define EOF_VALUE WEOF
#else
# define EOF_VALUE EOF
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MIN(a, b) ((a) < (b) ? (a) : (b))

typedef enum _CommandType {
  CommandTypeFill,
  CommandTypeRead,
  CommandTypeWrite
} CommandType;

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

static void
memcpyWithWidth(
  uint8_t* pDst, /* Might not be aligned */
  uint8_t* pSrc, /* Might not be aligned */
  size_t count, /* Might not be a multiple of width */
  unsigned int width) /* Preferred load/store width, in bytes */
{
  unsigned int alignment = ((unsigned int)(uintptr_t)pSrc) & (width - 1);
  size_t n = count / width;
  unsigned int preamble, postamble;
  
  alignment = ((unsigned int)(uintptr_t)pSrc) & (width - 1);
  if (alignment) {
    /* Use memcpy for unaligned part */
    preamble = width - alignment;
    if (preamble > count) {
      preamble = (unsigned int)count; /* Cast safe because preamble > count */
    }
    memcpy(pDst, pSrc, preamble);
    pDst += preamble;
    pSrc += preamble;
    count -= preamble;
  }
  n = count / width;
  
  /* Now that pSrc is aligned, can use 'width'-sized reads */
  switch (width) {
  case 1U:
    while (n--) {
      *pDst++ = *pSrc++;
    }
    break;
    
  case 2U:
    while (n--) {
      uint16_t val16 = *(uint16_t*)pSrc;
      pSrc += 2;
      val16 = cpu_to_le32(val16);
      *pDst++ = (uint8_t)((val16 >> 0) & 0xFFU);
      *pDst++ = (uint8_t)((val16 >> 8) & 0xFFU);
    }
    break;
    
  case 4U:
    while (n--) {
      uint32_t val32 = *(uint32_t*)pSrc;
      pSrc += 4;
      val32 = cpu_to_le32(val32);
      *pDst++ = (uint8_t)((val32 >> 0) & 0xFFU);
      *pDst++ = (uint8_t)((val32 >> 8) & 0xFFU);
      *pDst++ = (uint8_t)((val32 >> 16) & 0xFFU);
      *pDst++ = (uint8_t)((val32 >> 24) & 0xFFU);
    }
    break;
    
  case 8U:
    while (n--) {
      uint64_t val64 = *(uint64_t*)pSrc;
      pSrc += 8;
      val64 = cpu_to_le64(val64);
      *pDst++ = (uint8_t)((val64 >> 0) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 8) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 16) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 24) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 32) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 40) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 48) & 0xFFU);
      *pDst++ = (uint8_t)((val64 >> 56) & 0xFFU);
    }
    break;
  }
  
  postamble = (unsigned int)count & (width - 1); /* Cast safe because of limited range of (width - 1) */
  if (postamble) {
    /* Use memcpy for residual part */
    memcpy(pDst, pSrc, postamble);
  }
}

static void
dumpMemory(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int width,
  bool_t bBigEndian,
  bool_t bAlign,
  void* pBuffer,
  size_t n,
  uint64_t windowOffset)
{
  static const unsigned int addressSize = sizeof(uint64_t);
  unsigned int addressStrLen = 3 + 2 * addressSize + (addressSize - 1) / 4;
  uint64_t limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint8_t* p8;
  TCHAR line[257];
  TCHAR* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten;
  unsigned int remaining;

  assert(indent < 128);
  assert(nColumn == 1 || nColumn == 2 || nColumn == 4 || nColumn == 8 || nColumn == 16 || nColumn == 32);
  assert(width == 1 || width == 2 || width == 4 || width == 8);

  rowBytes = nColumn * width;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  limit = windowOffset + n;
  if (bAlign) {
    addressCol = windowOffset & ~(uint64_t)columnMask;
    addressLo = ((unsigned int)windowOffset) & columnMask;
  } else {
    addressCol = windowOffset;
    addressLo = 0;
  }

  m = limit - windowOffset;

  pPos = line;
  pPos[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent);
  pPos += MIN(indent, remaining);
  remaining -= MIN(indent, remaining);
  nwritten = _stprintf_s(pPos, remaining, _T("Dump of memory at 0x%08lX_%08lX + %llu(0x%llX) bytes:\n"),
    (unsigned long)(uint32_t)(windowOffset >> 32U), (unsigned long)(uint32_t)windowOffset,
    (unsigned long long)n, (unsigned long long)n);
  if (nwritten > 0) {
    pPos += nwritten;
    remaining -= nwritten;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 1 + addressStrLen);
  pPos += MIN(indent + 1 + addressStrLen, remaining);
  remaining -= MIN(indent + 1 + addressStrLen, remaining);
  for (i = 0; i < rowBytes; i += width) {
    switch (width) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lX "), (unsigned long)i);
      break;

    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lX "), (unsigned long)i);
      break;

    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lX "), (unsigned long)i);
      break;

    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lX "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
  }
  if (remaining) {
    *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
    *pPos++ = _T('\0');
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }

    memset(columnBuf, 0, rowBytes);
    memcpyWithWidth(columnBuf + addressLo, (uint8_t*)pBuffer, chunk, width);
    pBuffer = (void*)((uint8_t*)pBuffer + chunk);

    p8 = (uint8_t*)columnBuf;

    pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
    pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
    nwritten = _stprintf_s(pPos, remaining, _T("0x%08lX_%08lX: "), (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += width) {
      switch (width) {
        case 0x1U:
          val8 = *p8++;
          if (i >= addressLo && i < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)val8);
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
          break;

        case 0x2U:
          if (bBigEndian) {
            val16 = (uint16_t)(((uint16_t)p8[0] << 8U) | ((uint16_t)p8[1] << 0U));
          } else {
            val16 = (uint16_t)(((uint16_t)p8[0] << 0U) | ((uint16_t)p8[1] << 8U));
          }
          p8 += 2;
          for (k = 2; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 3) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;

        case 0x4U:
          if (bBigEndian) {
            val32 = (uint32_t)(((uint32_t)p8[0] << 24U) | ((uint32_t)p8[1] << 16U) | ((uint32_t)p8[2] <<  8U) | ((uint32_t)p8[3] <<  0U));
          } else {
            val32 = (uint32_t)(((uint32_t)p8[0] <<  0U) | ((uint32_t)p8[1] <<  8U) | ((uint32_t)p8[2] << 16U) | ((uint32_t)p8[3] << 24U));
          }
          p8 += 4;
          for (k = 4; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
          }
          break;

        case 0x8U:
          if (bBigEndian) {
            val64  = (uint64_t)(((uint64_t)p8[0] << 56U) | ((uint64_t)p8[1] << 48U) | ((uint64_t)p8[2] << 40U) | ((uint64_t)p8[3] << 32U));
            val64 |= (uint64_t)(((uint64_t)p8[4] << 24U) | ((uint64_t)p8[5] << 16U) | ((uint64_t)p8[6] <<  8U) | ((uint64_t)p8[7] <<  0U));
          } else {
            val64  = (uint64_t)(((uint64_t)p8[0] <<  0U) | ((uint64_t)p8[1] <<  8U) | ((uint64_t)p8[2] << 16U) | ((uint64_t)p8[3] << 24U));
            val64 |= (uint64_t)(((uint64_t)p8[4] << 32U) | ((uint64_t)p8[5] << 40U) | ((uint64_t)p8[6] << 48U) | ((uint64_t)p8[7] << 56U));
          }
          p8 += 8;
          for (k = 8; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 4) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;
      }
      if (remaining) {
        *pPos++ = _T(' ');
        remaining--;
      }
    }

    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;

      if (remaining) {
        if (_istprint(c) && c != _T('\t')) {
          *pPos++ = c;
        } else {
          *pPos++ = _T('.');
        }
        remaining--;
      }
    }
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    windowOffset += chunk;
  }
}

static int
fillMemoryUsingDS(
  unsigned int width,
  bool_t bBigEndian,
  ADMXRC3_HANDLE hDevice,
  unsigned int windowIndex,
  uint64_t windowOffset,
  unsigned int count,
  uint64_t fillValue)
{
  ADMXRC3_STATUS status;
  void* pWindow = NULL;
  uint64_t pos;
  unsigned int alignment;
  int ret = EXIT_OK; /* Assume success */
  
  status = ADMXRC3_MapWindow(hDevice, windowIndex, windowOffset, count, &pWindow);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to map %u(0x%X) bytes of window %u at offset %llu(0x%llX)\n"),
      count, count,
      (unsigned int)windowIndex,
      (unsigned long long)windowOffset, (unsigned long long)windowOffset);
    ret = EXIT_MAPWINDOW_ERROR;
    goto done;
  }
  
  alignment = ((unsigned int)(uintptr_t)pWindow) & (width - 1);
  if (0 == alignment) {
    switch (width) {
    case 1:
      if (bBigEndian) {
        for (pos = 0; pos < count; pos++) {
          *((uint8_t*)pWindow + pos) = cpu_to_be8((uint8_t)fillValue);
        }
      } else {
        for (pos = 0; pos < count; pos++) {
          *((uint8_t*)pWindow + pos) = cpu_to_le8((uint8_t)fillValue);
        }
      }
      break;

    case 2:
      if (bBigEndian) {
        for (pos = 0; pos < count; pos += 2) {
          *((uint16_t*)((uint8_t*)pWindow + pos)) = cpu_to_be16((uint16_t)fillValue);
        }
      } else {
        for (pos = 0; pos < count; pos += 2) {
          *((uint16_t*)((uint8_t*)pWindow + pos)) = cpu_to_le16((uint16_t)fillValue);
        }
      }
      break;

    case 4:
      if (bBigEndian) {
        for (pos = 0; pos < count; pos += 4) {
          *((uint32_t*)((uint8_t*)pWindow + pos)) = cpu_to_be32((uint32_t)fillValue);
        }
      } else {
        for (pos = 0; pos < count; pos += 4) {
          *((uint32_t*)((uint8_t*)pWindow + pos)) = cpu_to_le32((uint32_t)fillValue);
        }
      }
      break;

    case 8:
      if (bBigEndian) {
        for (pos = 0; pos < count; pos += 8) {
          *((uint64_t*)((uint8_t*)pWindow + pos)) = cpu_to_be64((uint64_t)fillValue);
        }
      } else {
        for (pos = 0; pos < count; pos += 8) {
          *((uint64_t*)((uint8_t*)pWindow + pos)) = cpu_to_le64((uint64_t)fillValue);
        }
      }
      break;

    default:
      assert(FALSE);
    }
  } else {
    uint8_t tmp8;
    uint16_t tmp16;
    uint32_t tmp32;
    uint64_t tmp64;
    
    _tprintf(_T("+++ Warning: unaligned virtual address %p; fill data will be written byte-by-byte.\n"), pWindow);
    
    switch (width) {
    case 1U:
      tmp8 = (uint8_t)(bBigEndian ? cpu_to_be8((uint8_t)fillValue) : cpu_to_le8((uint8_t)fillValue));
      for (pos = 0; pos < count;) {
        *((uint8_t*)pWindow + pos++) = tmp8;
      }
      break;
    
    case 2U:
      tmp16 = (uint16_t)(bBigEndian ? cpu_to_be16((uint16_t)fillValue) : cpu_to_le16((uint16_t)fillValue));
      for (pos = 0; pos < count;) {
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp16 >> 0);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp16 >> 8);
      }
      break;
    
    case 4U:
      tmp32 = (uint32_t)(bBigEndian ? cpu_to_be32((uint32_t)fillValue) : cpu_to_le32((uint32_t)fillValue));
      for (pos = 0; pos < count;) {
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp32 >> 0);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp32 >> 8);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp32 >> 16);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp32 >> 24);
      }
      break;
    
    case 8U:
      tmp64 = (uint64_t)(bBigEndian ? cpu_to_be64((uint64_t)fillValue) : cpu_to_le64((uint64_t)fillValue));
      for (pos = 0; pos < count; pos += 8) {
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 0);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 8);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 16);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 24);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 32);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 40);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 48);
        *((uint8_t*)pWindow + pos++) = (uint8_t)(tmp64 >> 56);
      }
      break;
    }
  }

done:
  if (NULL != pWindow) {
    ADMXRC3_UnmapWindow(hDevice, pWindow);
  }
  
  return ret;
}

static int
writeDataUsingDS(
  unsigned int width,
  ADMXRC3_HANDLE hDevice,
  unsigned int windowIndex,
  uint64_t windowOffset,
  void* pDataBlock,
  unsigned int count)
{
  ADMXRC3_STATUS status;
  void* pWindow = NULL;
  uint64_t pos;
  unsigned int alignment;
  int ret = EXIT_OK; /* Assume success */
  
  status = ADMXRC3_MapWindow(hDevice, windowIndex, windowOffset, count, &pWindow);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to map %u(0x%X) bytes of window %u at offset %llu(0x%llX)\n"),
      count, count,
      (unsigned int)windowIndex,
      (unsigned long long)windowOffset, (unsigned long long)windowOffset);
    ret = EXIT_MAPWINDOW_ERROR;
    goto done;
  }
  
  alignment = ((unsigned int)(uintptr_t)pWindow) & (width - 1);
  if (0 == alignment) {
    switch (width) {
    case 1:
      for (pos = 0; pos < count; pos += sizeof(uint8_t)) {
        *((uint8_t*)pWindow + pos) = *((uint8_t*)pDataBlock + pos);
      }
      break;

    case 2:
      for (pos = 0; pos < count; pos += sizeof(uint16_t)) {
        *(uint16_t*)((uint8_t*)pWindow + pos) = *(uint16_t*)((uint8_t*)pDataBlock + pos);
      }
      break;

    case 4:
      for (pos = 0; pos < count; pos += sizeof(uint32_t)) {
        *(uint32_t*)((uint8_t*)pWindow + pos) = *(uint32_t*)((uint8_t*)pDataBlock + pos);
      }
      break;

    case 8:
      for (pos = 0; pos < count; pos += sizeof(uint64_t)) {
        *(uint64_t*)((uint8_t*)pWindow + pos) = *(uint64_t*)((uint8_t*)pDataBlock + pos);
      }
      break;

    default:
      assert(FALSE);
    }
  } else {
    _tprintf(_T("+++ Warning: unaligned virtual address %p; data will be written using memcpy().\n"), pWindow);
    
    memcpy(pWindow, pDataBlock, count);
  }

done:
  if (NULL != pWindow) {
    ADMXRC3_UnmapWindow(hDevice, pWindow);
  }
  
  return ret;
}

static int
dumpMemoryUsingDS(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int width,
  bool_t bBigEndian,
  bool_t bAlign,
  ADMXRC3_HANDLE hDevice,
  unsigned int windowIndex,
  unsigned int count,
  uint64_t windowOffset)
{
  ADMXRC3_STATUS status;
  void* pWindow = NULL;
  unsigned int alignment;
  int ret = EXIT_OK; /* Assume success */
  
  status = ADMXRC3_MapWindow(hDevice, windowIndex, windowOffset, count, &pWindow);
  if (ADMXRC3_SUCCESS != status) {
    _tprintf(_T("*** Failed to map %u(0x%X) bytes of window %u at offset %llu(0x%llX)\n"),
      count, count,
      (unsigned int)windowIndex,
      (unsigned long long)windowOffset, (unsigned long long)windowOffset);
    ret = EXIT_MAPWINDOW_ERROR;
    goto done;
  }
  
  alignment = ((unsigned int)(uintptr_t)pWindow) & (width - 1);
  if (0 != alignment) {
    _tprintf(_T("+++ Warning: unaligned virtual address %p; data will be read byte-by-byte.\n"), pWindow);
  }
  
  dumpMemory(indent, nColumn, width, bBigEndian, bAlign, pWindow, count, windowOffset);
  
done:
  if (NULL != pWindow) {
    ADMXRC3_UnmapWindow(hDevice, pWindow);
  }
  return ret;
}

#if (defined(__VXWORKS__) || defined(__vxworks))
 
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

#define FLAG_BYSERIAL   (0x1U) /* Interpret 'indexOrSerial' as a serial number */
#define FLAG_BIGENDIAN (0x10U) /* Read and write values in big-endian byte order */

static int
admxrc3Fill(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  uint64_t windowOffset,
  unsigned int count,
  unsigned int width,
  uint64_t value)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  unsigned int alignment;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  alignment = count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  ret = fillMemoryUsingDS(width, bBigEndian, hDevice, windowIndex, windowOffset, count, value);
  if (ret) {
    goto done;
  }

done:
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

int
admxrc3DumpHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3DumpHelp()\n"));
  _tprintf(_T("  int admxrc3FB (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, val8)\n"));
  _tprintf(_T("  int admxrc3XFB(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, val8)\n"));
  _tprintf(_T("  int admxrc3FW (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, val16)\n"));
  _tprintf(_T("  int admxrc3XFW(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, val16)\n"));
  _tprintf(_T("  int admxrc3FD (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, val32)\n"));
  _tprintf(_T("  int admxrc3XFD(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, val32)\n"));
  _tprintf(_T("  int admxrc3FQ (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, val64(ULL))\n"));
  _tprintf(_T("  int admxrc3XFQ(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, val64(ULL))\n"));
  _tprintf(_T("  int admxrc3RB (indexOrSerial, flags, windowIndex, windowOffset,      byteCount)\n"));
  _tprintf(_T("  int admxrc3XRB(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3RW (indexOrSerial, flags, windowIndex, windowOffset,      byteCount)\n"));
  _tprintf(_T("  int admxrc3XRW(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3RD (indexOrSerial, flags, windowIndex, windowOffset,      byteCount)\n"));
  _tprintf(_T("  int admxrc3XRD(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3RQ (indexOrSerial, flags, windowIndex, windowOffset,      byteCount)\n"));
  _tprintf(_T("  int admxrc3XRQ(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount)\n"));
  _tprintf(_T("  int admxrc3WB (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3XWB(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3WW (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3XWW(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3WD (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3XWD(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, ...)\n"));
  _tprintf(_T("  int admxrc3WQ (indexOrSerial, flags, windowIndex, windowOffset,      byteCount, ...(ULL))\n"));
  _tprintf(_T("  int admxrc3XWQ(indexOrSerial, flags, windowIndex, windowOffset(ULL), byteCount, ...(ULL))\n"));
  _tprintf(_T("'localAddress(ULL)' & 'val64(ULL)' are unsigned long long parameters.\n"));
  _tprintf(_T("'...(ULL)' is a variable number of unsigned long long parameters.\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL  0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_BIGENDIAN 0x%02X => Loads and stores use big-endian byte-order\n"), FLAG_BIGENDIAN);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3XFB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  unsigned int val8)
{
  return admxrc3Fill(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 1, (uint64_t)val8);
}

int
admxrc3FB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  unsigned int val8)
{
  return admxrc3XFB(indexOrSerial, flags, windowIndex, windowOffset, count, val8);
}

int
admxrc3XFW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  unsigned int val16)
{
  return admxrc3Fill(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 2, (uint64_t)val16);
}

int
admxrc3FW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  unsigned int val16)
{
  return admxrc3XFW(indexOrSerial, flags, windowIndex, windowOffset, count, val16);
}

int
admxrc3XFD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  unsigned int val32)
{
  return admxrc3Fill(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 4, (uint64_t)val32);
}

int
admxrc3FD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  unsigned int val32)
{
  return admxrc3XFD(indexOrSerial, flags, windowIndex, windowOffset, count, val32);
}

int
admxrc3XFQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  unsigned long long val64)
{
  return admxrc3Fill(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 8, (uint64_t)val64);
}

int
admxrc3FQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  unsigned long long val64)
{
  return admxrc3XFQ(indexOrSerial, flags, windowIndex, windowOffset, count, val64);
}

static int
admxrc3Dump(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  uint64_t windowOffset,
  unsigned int count,
  unsigned int width)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  unsigned int alignment;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  alignment = count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  ret = dumpMemoryUsingDS(0, 16 / width, width, bBigEndian, TRUE, hDevice, windowIndex, count, windowOffset);
  if (ret) {
    goto done;
  }

done:
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

int
admxrc3XRB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count)
{
  return admxrc3Dump(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 1);
}

int
admxrc3RB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count)
{
  return admxrc3XRB(indexOrSerial, flags, windowIndex, windowOffset, count);
}

int
admxrc3XRW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count)
{
  return admxrc3Dump(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 2);
}

int
admxrc3RW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count)
{
  return admxrc3XRW(indexOrSerial, flags, windowIndex, windowOffset, count);
}

int
admxrc3XRD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count)
{
  return admxrc3Dump(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 4);
}

int
admxrc3RD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count)
{
  return admxrc3XRD(indexOrSerial, flags, windowIndex, windowOffset, count);
}

int
admxrc3XRQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count)
{
  return admxrc3Dump(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 8);
}

int
admxrc3RQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count)
{
  return admxrc3XRQ(indexOrSerial, flags, windowIndex, windowOffset, count);
}

static int
writeArgList(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  uint64_t windowOffset,
  unsigned int count,
  unsigned int width,
  va_list args)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  bool_t bOpenBySerial = FALSE;
  bool_t bBigEndian = FALSE;
  uint8_t* pDataBlock = NULL;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int alignment;
  size_t pos;
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial  = TRUE;
  }
  if (flags & FLAG_BIGENDIAN) {
    bBigEndian  = TRUE;
  }
  
  alignment = count & (width - 1U);
  if (alignment) {
    _tprintf(_T("+++ Warning: byte count %u(0x%X) is not an integer multiple of width (%u)\n"), count, count, width);
    count -= alignment;
    _tprintf(_T("+++          byte count adjusted to %u(0x%X)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }
  
  pDataBlock = (uint8_t*)malloc(count);
  if (NULL == pDataBlock) {
    _tprintf(_T("*** Failed to allocate %u(0x%X) bytes for fill data buffer."),
      count, count);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  status = sampleOpenCardNoClearError(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  pos = 0;
  while (pos < count) {
    switch (width) {
    case 1:
      val8 = (uint8_t)va_arg(args, unsigned int);
      *(pDataBlock + pos) = (uint8_t)(bBigEndian ? cpu_to_be8(val8) : cpu_to_le8(val8));
      pos += 1;
      break;
      
    case 2:
      val16 = (uint16_t)va_arg(args, unsigned int);
      *((uint16_t*)(pDataBlock + pos)) = (uint16_t)(bBigEndian ? cpu_to_be16(val16) : cpu_to_le16(val16));
      pos += 2;
      break;
      
    case 4:
      val32 = (uint32_t)va_arg(args, unsigned int);
      *((uint32_t*)(pDataBlock + pos)) = (uint32_t)(bBigEndian ? cpu_to_be32(val32) : cpu_to_le32(val32));
      pos += 4;
      break;
      
    case 8:
      val64 = (uint64_t)va_arg(args, unsigned long long);
      *((uint64_t*)(pDataBlock + pos)) = (uint64_t)(bBigEndian ? cpu_to_be64(val64) : cpu_to_le64(val64));
      pos += 8;
      break;
    }
  }
  
  ret = writeDataUsingDS(width, hDevice, windowIndex, windowOffset, pDataBlock, count);
  if (ret) {
    goto done;
  }

done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

int
admxrc3XWB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 1, args);
  va_end(args);
  
  return ret;
}

int
admxrc3WB(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 1, args);
  va_end(args);
  
  return ret;
}

int
admxrc3XWW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 2, args);
  va_end(args);
  
  return ret;
}

int
admxrc3WW(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 2, args);
  va_end(args);
  
  return ret;
}

int
admxrc3XWD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 4, args);
  va_end(args);
  
  return ret;
}

int
admxrc3WD(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 4, args);
  va_end(args);
  
  return ret;
}

int
admxrc3XWQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned long long windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 8, args);
  va_end(args);
  
  return ret;
}

int
admxrc3WQ(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int windowIndex,
  unsigned int windowOffset,
  unsigned int count,
  ...)
{
  va_list args;
  int ret;

  va_start(args, count);
  ret = writeArgList(indexOrSerial, flags, windowIndex, (uint64_t)windowOffset, count, 8, args);
  va_end(args);
  
  return ret;
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  unsigned int tokenFlags = 0;
  Option options[4];
  Arguments arguments = g_initArguments;
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  bool_t bBigEndian, bInterpretHex;
  CommandType command;
  unsigned int count, width, index, serial, windowIndex;
  uint64_t windowOffset;
  TCHAR* pCommand;
  uint8_t* pDataBlock = NULL;
  int ret = EXIT_OK;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("be");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("TRUE => use big-endian memory format");
  options[3].key = _T("hex");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("TRUE => interpret values as being hexadecimal");
  arguments.nOption = 4;
  arguments.option = options;
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("Command to perform: fb, fw, fd, fq, rb, rw, rd, rq, wb, ww, wd or wq");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("window");
  paramMeaning[1] = _T("Index of window to use");
  paramType[2] = OptionUInt64;
  paramInfo[2] = _T("address");
  paramMeaning[2] = _T("Local / OCP address to fill, read or write");
  paramType[3] = OptionUInt;
  paramInfo[3] = _T("length");
  paramMeaning[3] = _T("Number of bytes to fill, read or write");
  paramType[4] = OptionString;
  paramInfo[4] = _T("item ...");
  paramMeaning[4] = _T("Fill value or data items to write");
  arguments.nParamType = 5;
  arguments.minNParam = 3;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;
  
  if (sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bBigEndian = options[2].value.booleanVal;
  bInterpretHex = options[3].value.booleanVal;
  if (bInterpretHex) {
    tokenFlags |= SAMPLE_TOKEN_FLAG_INTERPRET_HEX;
  }
  pCommand = arguments.param[0].pStringVal;
  windowIndex = arguments.param[1].uintVal;
  windowOffset = arguments.param[2].uint64Val;

  if (_tcsicmp(pCommand, _T("fb")) == 0) {
    width = 1;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fw")) == 0) {
    width = 2;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fd")) == 0) {
    width = 4;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fq")) == 0) {
    width = 8;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    command = CommandTypeWrite;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = EXIT_UNRECOGNIZED_COMMAND;
    goto done;
  }

  if (arguments.nParam < 4) {
    count = width;
  } else {
    count = arguments.param[3].uintVal;
  }
  if (count % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n"), count, count, width);
    count &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%x)\n"), count, count);
  }
  if (count == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  if (command == CommandTypeFill) {
    SampleTokenState state = g_sampleTokenStateInit;
    
    /* Attempt to parse 5th positional argument */
    if (arguments.nParam >= 5) {
      sampleTokenPumpString(&state, arguments.param[4].pStringVal, tokenFlags, NULL);
    }
    
    if (!sampleTokenParsed(&state)) {
      /* No 5th positional argument, so attempt to get token from stdin */
      _tprintf(_T("Enter fill value: "));
      fflush(stdout);
      state = g_sampleTokenStateInit;
      if (!feof(stdin)) {
        sampleTokenPumpFile(&state, stdin, tokenFlags);
      }
    }
    
    if (!sampleTokenParsed(&state)) {
      /* No token obtained, so do nothing */
      goto done;
    }
    
    ret = fillMemoryUsingDS(width, bBigEndian, hDevice, windowIndex, windowOffset, count, (uint64_t)state.result.value);
    if (ret) {
      goto done;
    }
  } else if (command == CommandTypeWrite) {
    SampleTokenState state;
    unsigned int pos;
    int i;

    pDataBlock = (uint8_t*)malloc(count);
    if (NULL == pDataBlock) {
      _tprintf(_T("*** Failed to allocate %u(%X) bytes to hold write values\n"), count, count);
      ret = EXIT_ALLOCATION_FAILURE;
      goto done;
    }
    
    i = 4;
    pos = 0;
    while (pos < count) {
      state = g_sampleTokenStateInit;
      
      /* Attempt to parse 5th and later positional arguments */
      if (i < arguments.nParam) {
        sampleTokenPumpString(&state, arguments.param[i].pStringVal, tokenFlags, NULL);
        i++;
        if (!sampleTokenParsed(&state)) {
          continue;
        }
        _tprintf(_T("0x%08lX_%08lX: 0x%llX\n"),
          (unsigned long)(uint32_t)((windowOffset + pos) >> 32), (unsigned long)(uint32_t)(windowOffset + pos),
          state.result.value);        
      }
      
      if (!sampleTokenParsed(&state)) {
        /* No positional arguments left, so attempt to get token from stdin */
        _tprintf(_T("0x%08lX_%08lX: "),
          (unsigned long)(uint32_t)((windowOffset + pos) >> 32), (unsigned long)(uint32_t)(windowOffset + pos));
        fflush(stdout);
        if (!feof(stdin)) {
          sampleTokenPumpFile(&state, stdin, tokenFlags);
        }          
      }
      
      if (!sampleTokenParsed(&state)) {
        /* No more tokens obtained, so quit from loop */
        break;
      }
      
      switch (width) {
      case 1:
        *(pDataBlock + pos) = bBigEndian ? cpu_to_be8((uint8_t)state.result.value) : cpu_to_le8((uint8_t)state.result.value);
        pos += 1;
        break;
        
      case 2:
        *((uint16_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be16((uint16_t)state.result.value) : cpu_to_le16((uint16_t)state.result.value);
        pos += 2;
        break;
        
      case 4:
        *((uint32_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be32((uint32_t)state.result.value) : cpu_to_le32((uint32_t)state.result.value);
        pos += 4;
        break;
        
      case 8:
        *((uint64_t*)(pDataBlock + pos)) = bBigEndian ? cpu_to_be64((uint64_t)state.result.value) : cpu_to_le64((uint64_t)state.result.value);
        pos += 8;
        break;
      }
    }
    
    count = pos;
    ret = writeDataUsingDS(width, hDevice, windowIndex, windowOffset, pDataBlock, count);
    if (ret) {
      goto done;
    }
  } else if (command == CommandTypeRead) {
    ret = dumpMemoryUsingDS(0, 16 / width, width, bBigEndian, TRUE, hDevice, windowIndex, count, windowOffset);
    if (ret) {
      goto done;
    }
  } else {
    /* Should never get here */
    assert(FALSE);
  }

done:
  if (NULL != pDataBlock) {
    free(pDataBlock);
  }
  
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
