/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC series card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility for:
**
**   o ADPe-XRC-5T
**   o ADM-XRC-6TL
**   o ADM-XRC-6T1
**   o ADM-XRC-6T-DA1
**   o ADM-XRC-6TGE & ADM-XRC-6TGEL
**   o ADM-XRC-6T-ADV8
**   o ADPE-XRC-6T & ADPE-XRC-6T-L
**   o ADPE-XRC-6T-ADV (Controller)
**   o ADM-XRC-7K1
**   o ADM-XRC-7V1
**   o ADM-VPX3-7V2
**   o ADM-PCIE-7V3
**   o ADM-XRC-KU1 & ADM-XRC-KU1A-T8
**   o ADM-PCIE-KU3
**   o ADM-PCIE-8V3
**   o ADM-PCIE-8K5
**
** (C) Copyright Alpha Data 2009-2016
**
** Exit codes: see EXIT_* #defines below.
**
** NOTE: This utility applies only to the regions marked (*) in the comments
**       below, and cannot affect the other regions.
**
**  1. The Flash address map used on the ADM-XRC-6TL, ADM-XRC-6T1,
**     ADM-XRC-6T-DA1, ADM-XRC-6TGE, ADPE-XRC-6T(-L), ADM-XRC-7K1 &
**     ADM-XRC-7V1 & ADM-VPX3-7V2 is:
**
**     Flash bank 0:
**     0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP Bridge bitstream
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP Bridge bitstream
**     0x1000000 - 0x10FFFFF  Alpha Data VPD region
**     0x1100000 - 0x11FFFFF  Customer VPD region
**     0x1200000 - 0x28FFFFF  Target FPGA bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream (*)
**
**  2. The Flash address map used on the ADM-XRC-6T-ADV8 is:
**
**     Flash bank 0:
**     0x0000000 - 0x1FFFFFF  Failsafe FPGA bitstream (*)
**     0x2000000 - 0x3FFFFFF  Normal FPGA bitstream (*)
**
**  3. The Flash address map used on the ADM-XRC-6T-ADV (Controller) is:
**
**     Flash bank 0 (U52):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     Flash bank 1 (U51):
**     0x0000000 - 0x0FFFFFF  Customer region
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 1 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 1 bitstream (*)
**
**  4. The Flash address map used on the ADM-PCIE-7V3 & ADM-PCIE-KU3 is:
**
**     Flash bank 0:
**     0x0000000 - 0x1FFFFFF  Failsafe FPGA bitstream (*)
**     0x2000000 - 0x3FFFFFF  Normal FPGA bitstream (*)
**     0x4000000 - 0x5FFFFFF  User-definable (*)
**     0x6000000 - 0x7FFFFFF  User-definable (*)
**
**  5. The Flash address map used on the ADM-XRC-KU1 is:
**
**     Flash bank 0:
**     0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP/AXI4 Bridge bitstream
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP/AXI4 Bridge bitstream
**     0x1000000 - 0x11FFFFF  (reserved)
**     0x1200000 - 0x4FFFFFF  Target FPGA bitstream (*)
**     0x5000000 - 0x7FFFFFF  Available for user data (*)
**
**  6. The Flash address map used on the ADM-PCIE-8V3 & ADM-PCIE-8K5 is:
**
**     Flash bank 0:
**     0x0000000 - 0x3FFFFFF  Failsafe FPGA bitstream (*)
**     0x4000000 - 0x7FFFFFF  Normal FPGA bitstream (*)
**
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcschr(x, c) strchr(x, c)
# define _tcslen(x) strlen(x)
# define _totlower(c) tolower(c)
# define _tmain main
# define _tprintf printf
# define _stscanf_s sscanf
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#ifdef _WIN32
# define PROGRAM_NAME _T("flash.exe")
#else
# define PROGRAM_NAME _T("flash")
#endif

/* Exit codes */
#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_BAD_RANGE               (2) /* Value for -range option is not valid */
#define EXIT_UNRECOGNIZED_COMMAND    (3) /* Command not recognized */
#define EXIT_INSUFFICIENT_ARGS       (4) /* Too few positional arguments following command */
#define EXIT_ILLEGAL_FLAGS           (5) /* An illegal combination of flags was passed */
#define EXIT_INVALID_RANGE           (6) /* A Flash address range was invalid */
#define EXIT_ALLOCATION_FAILURE      (7) /* A memory allocation failed */
#define EXIT_BLOCK_QUERY_FAILURE     (8) /* Failed to query a Flash block (ADMXRC3_GetFlashBlockInfo failed) */
#define EXIT_CREATE_FILE_ERROR       (9) /* Failed to open file for writing readback data */
#define EXIT_WRITE_FILE_ERROR       (10) /* Failed to write to readback data file */
#define EXIT_READ_FLASH_ERROR       (11) /* Failed to read from Flash (ADMXRC3_ReadFlash failed) */
#define EXIT_WRITE_FLASH_ERROR      (12) /* Failed to write to Flash (ADMXRC3_WriteFlash failed) */
#define EXIT_ERASE_FLASH_ERROR      (13) /* Failed to erase Flash (ADMXRC3_EraseFlash failed) */
#define EXIT_READ_BIT_ERROR         (14) /* Failed to read .bit file */
#define EXIT_SANITY_ERROR           (15) /* A sanity check failed */
#define EXIT_FPGA_MISMATCH          (16) /* The FPGA identifier in the .bit file does not match the FPGA in the device */
#define EXIT_BOOT_FLAG_VERIFY_ERROR (17) /* Boot flag read back has incorrect value */
#define EXIT_DEVICE_OPEN_ERROR      (18) /* Failed to open ADMXRC3 device */
#define EXIT_CARDINFO_ERROR         (19) /* Failed to get information about ADMXRC3 device */
#define EXIT_FAILSAFE_NOT_SUPPORTED (20) /* Writing to failsafe region not supported for this ADMXRC3 device */
#define EXIT_ILLEGAL_TARGET_INDEX   (21) /* Target FPGA index illegal for this ADMXRC3 device */
#define EXIT_UNSUPPORTED_MODEL      (22) /* The ADMXRC3 device is an unsupported model */
#define EXIT_FPGAINFO_ERROR         (23) /* Failed to get information about target FPGA */
#define EXIT_FLASHINFO_ERROR        (24) /* Failed to get information about Flash memory */
#define EXIT_ILLEGAL_REGION         (25) /* Region index passed is illegal for this ADMXRC3 device */
#define EXIT_SET_CLOCK_FAILED       (26) /* Failed to set LCLK clock generator to nominal frequency */
#define EXIT_NOT_BLANK              (27) /* Blank check failed; region is not blank */
#define EXIT_VERIFY_FAILED          (28) /* Verification failed; data read back did not match data written */
#define EXIT_NULL_FILENAME          (29) /* Illegal NULL string passed for a filename */

#define BLANK_CHECK_BUFFER_SIZE (0x10000)
#define READBACK_BUFFER_SIZE (0x10000)
#define VERIFY_BUFFER_SIZE (0x10000)

const uint64_t sizeTMax = (uint64_t)(size_t)-1;

typedef enum _CommandType {
  CommandTypeInfo,
  CommandTypeProgram,
  CommandTypeVerify,
  CommandTypeErase,
  CommandTypeChkblank,
  CommandTypeReadback
} CommandType;

typedef enum _BootFlagType {
  /* ADM-XRC-6T-ADV8 */
  BootFlagTypeNone,          /* No boot flag is programmed into Flash */

  /* ADM-XRC-II */
  BootFlagType8Bit,          /* [7:0] = 0 */

  /*
  ** ADPe-XRC-5T, ADM-XRC-6TL, ADM-XRC-6T1, ADM-XRC-6T-DA1, ADM-XRC-6TGE,
  ** ADM-XRC-6T-ADV8, ADPE-XRC-6T(-L), ADPE-XRC-6T-ADV, ADM-XRC-7K1,
  ** ADM-XRC-7V1
  */
  BootFlagType24BitAddress,  /* [31:8] = length in bytes - 1, [7:0] = 0 */

  /* ADM-VPX3-7V2, ADM-XRC-KU1 */
  BootFlagType30BitAddress,  /* [31] = 0, [30] = (reserved), [29:0] = length in bytes - 1 */
} BootFlagType;

typedef enum _SwapFlags {
  SwapFlagsNone      = 0,
  SwapFlagsBitSwap   = 0x1 << 0,    /* Do bit-swapping inside bytes */
  SwapFlagsByteSwap2 = 0x1 << 1,    /* Do 2-byte swapping; mutually exclusive with SwapFlagsByteSwap4 */
  SwapFlagsByteSwap4 = 0x2 << 1,    /* Do 4-byte swapping; mutually exclusive with SwapFlagsByteSwap2 */
} SwapFlags;

typedef struct _Region {
  uint64_t base;
  uint64_t limit;
  uint64_t size;
} Region;

#define EMPTY_REGION { 0, 0, 0 }

static const SwapFlags SwapFlagsByteSwapMask = (SwapFlags)(0x3 << 1);

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

/*
** Returns TRUE if device identifier in bitstream matches device identifier
** in target FPGA information, otherwise FALSE.
**
** To handle the fact that identifiers from Ultrascale bitstreams, unlike
** previous FPGA families, have the speed and temperature grade as suffices,
** consider the identifiers to match iff the identifier from target FPGA
** information is a prefix of the identifier from the bitstream.
*/
static bool_t
checkDeviceMatch(
  const TCHAR* pBitstreamIdentifier,
  const TCHAR* pFPGAIdentifier)
{
  size_t n = _tcslen(pBitstreamIdentifier);
  size_t m = _tcslen(pFPGAIdentifier);
  size_t i;

  if (n < m) {
    /* If FPGA identifier is shorter than bitstream identifier, one or the other must be truncated or malformed, so consider not matched. */
    return FALSE;
  }

  for (i = 0; i < m; i++) {
    TCHAR c1 = (TCHAR)_totlower(pBitstreamIdentifier[i]);
    TCHAR c2 = (TCHAR)_totlower(pFPGAIdentifier[i]);
    if (c1 != c2) {
      return FALSE;
    }
  }

  return TRUE;
}

/*
** Establish defaults, according to model, for:
**
**  o The default region used for erase/program/verify/chkblank commands.
**  o The failsafe region (for +failsafe option).
**  o The number of regions.
**  o The type of boot flag, if any.
**  o The length of the boot flag, if any.
**  o Any bit- or byte-swapping that needs to be performed on the SelectMap data.
*/
static void
getModelDefaults(
  ADMXRC3_MODEL_TYPE model,
  unsigned int* pDefaultRegion,
  unsigned int* pFailsafeRegion,
  unsigned int* pNumRegion,
  BootFlagType* pBootFlagType,
  SwapFlags* pSwapFlags)
{
  unsigned int defaultRegion = 0U;
  unsigned int failsafeRegion = 0U;
  unsigned int numRegion = 0U;
  BootFlagType bootFlagType = BootFlagTypeNone;
  SwapFlags swapFlags = SwapFlagsNone;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    defaultRegion = 0;
    numRegion = 1;
    bootFlagType = BootFlagType8Bit;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
    defaultRegion = 0;
    failsafeRegion = 1;
    numRegion = 2;
    bootFlagType = BootFlagType24BitAddress;
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 2;
    bootFlagType = BootFlagTypeNone;
    swapFlags = (SwapFlags)(SwapFlagsBitSwap | SwapFlagsByteSwap2);
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    defaultRegion = 0;
    failsafeRegion = 1;
    numRegion = 2;
    bootFlagType = BootFlagType24BitAddress;
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    defaultRegion = 0;
    numRegion = 1;
    bootFlagType = BootFlagType30BitAddress;
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 4;
    bootFlagType = BootFlagTypeNone;
    swapFlags = (SwapFlags)(SwapFlagsBitSwap | SwapFlagsByteSwap2);
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    defaultRegion = 0;
    numRegion = 1;
    bootFlagType = BootFlagType30BitAddress;
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 2;
    bootFlagType = BootFlagTypeNone;
    swapFlags = (SwapFlags)(SwapFlagsBitSwap | SwapFlagsByteSwap2);
    break;

  default:
    assert(FALSE);
    break;
  }

  *pDefaultRegion = defaultRegion;
  *pFailsafeRegion = failsafeRegion;
  *pNumRegion = numRegion;
  *pBootFlagType = bootFlagType;
  *pSwapFlags = swapFlags;
}

/*
** Figure out the range of Flash addresses for a particular region.
** The address range depends upon the region and the model.
*/
static void
getEraseAddressRange(
  ADMXRC3_MODEL_TYPE model,
  const ADMXRC3_FLASH_INFO* pFlashInfo,
  unsigned int regionIndex,
  Region* pRegion)
{
  uint64_t eraseAddress = 0U;
  uint64_t eraseLimit = 0U;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    switch (regionIndex) {
    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + pFlashInfo->UseableLength;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    switch (regionIndex) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0xC00000U;
      eraseLimit = pFlashInfo->UseableStart + 0x1800000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0xC00000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
    switch (regionIndex) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x1700000U;
      eraseLimit = pFlashInfo->UseableStart + 0x2E00000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x1700000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    switch (regionIndex) {
    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x2000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x2000000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    switch (regionIndex) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x1700000U;
      eraseLimit = pFlashInfo->UseableStart + 0x2E00000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x1700000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    /* Default */
    eraseAddress = pFlashInfo->UseableStart;
    eraseLimit = eraseAddress + pFlashInfo->UseableLength;
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
    switch (regionIndex) {
    case 3:
      /* User-definable */
      eraseAddress = pFlashInfo->UseableStart + 0x6000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x8000000U;
      break;

    case 2:
      /* User-definable */
      eraseAddress = pFlashInfo->UseableStart + 0x4000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x6000000U;
      break;

    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x2000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x2000000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    switch (regionIndex) {
    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x4000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x8000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    switch (regionIndex) {
    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x3E00000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  default:
    assert(FALSE);
    break;
  }

  pRegion->base = eraseAddress;
  pRegion->limit = eraseLimit;
  pRegion->size = eraseLimit - eraseAddress;
}

/*
** Figure out the address at which to program the SelectMap data, and the
** address of the boot flag (if there is one). This is calculated as an
** offset, which depends upon the model, from the starting erase address.
*/
static void
getProgramAddress(
  ADMXRC3_MODEL_TYPE model,
  uint64_t eraseAddress,
  BootFlagType bootFlagType,
  Region* pBootFlagRegion,
  uint64_t* pProgramAddress)
{
  uint64_t bootFlagAddress = 0U;
  uint64_t programAddress = 0U;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    bootFlagAddress = eraseAddress + 0x8000U;
    programAddress = eraseAddress + 0x8001U;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMVPX37V2:
  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    bootFlagAddress = eraseAddress;
    programAddress = eraseAddress + 0x20000U;
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    programAddress = eraseAddress;
    break;

  default:
    assert(FALSE);
    break;
  }

  pBootFlagRegion->base = bootFlagAddress;
  switch (bootFlagType) {
  case BootFlagTypeNone:
    pBootFlagRegion->size = 0;
    break;

  case BootFlagType8Bit:
    pBootFlagRegion->size = 1;
    break;

  case BootFlagType24BitAddress:
  case BootFlagType30BitAddress:
    pBootFlagRegion->size = 4;
    break;

  default:
    assert(FALSE);
  }
  pBootFlagRegion->limit = pBootFlagRegion->base + pBootFlagRegion->size;

  *pProgramAddress = programAddress;
}

static int
eraseFlash(
  ADMXRC3_HANDLE hDevice,
  unsigned int flashIndex,
  uint64_t eraseStart,
  uint64_t eraseLimit)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status;
  uint64_t address, remaining;
  int ret = EXIT_OK;
  
  address = eraseStart;
  remaining = eraseLimit - eraseStart;
  while (address < eraseLimit) {
    status = ADMXRC3_GetFlashBlockInfo(hDevice, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_BLOCK_QUERY_FAILURE;
      goto done;
    }

    /*
    ** Adjust blockInfo address & length so that it does not erase anything
    ** outside of the address range [eraseStart, eraseLimit).
    */
    if (blockInfo.Address < address) {
      uint64_t difference = address - blockInfo.Address;
      blockInfo.Address = address;
      blockInfo.Length -= difference;
    }
    if (blockInfo.Length > remaining) {
      blockInfo.Length = remaining;
    }

    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_EraseFlash(hDevice, flashIndex, 0, blockInfo.Address, blockInfo.Length);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to erase Flash at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)blockInfo.Address, (unsigned long long)(blockInfo.Address + blockInfo.Length - 1U), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_ERASE_FLASH_ERROR;
      goto done;
    }

    /* Advance to next block */
    address = blockInfo.Address + blockInfo.Length;
    remaining = remaining - blockInfo.Length;
    if (address <= eraseStart) {
      /* Catch 'address' wrapping around to 0 */
      break;
    }
  }

done:
  return ret;
}

static int
programBitstream(
  ADMXRC3_HANDLE hDevice,
  ADMXRC3_BITSTREAM* pBitstream,
  unsigned int flashIndex,
  uint64_t programStart)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status;
  uint8_t* pData;
  uint64_t address, chunk;
  size_t remaining;
  int ret = EXIT_OK;

  address = programStart;
  remaining = pBitstream->Length;
  pData = &pBitstream->Data[0];
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hDevice, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_BLOCK_QUERY_FAILURE;
      goto done;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_WriteFlash(hDevice, flashIndex, 0, address, (size_t)chunk, pData); /* Cast is safe because we guarantee chunk <= maximum size_t value */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to write Flash at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_WRITE_FLASH_ERROR;
      goto done;
    }

    address += chunk;
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
    pData += (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
  }

done:
  return ret;
}

static int
checkBlank(
  ADMXRC3_HANDLE hDevice,
  unsigned int flashIndex,
  uint64_t checkStart,
  uint64_t checkLimit,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, nError = 0, i, chunk;
  int ret = EXIT_OK;
  
  pBuffer = (uint8_t*)malloc(BLANK_CHECK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate blank check buffer of size 0x%lx bytes\n"),
      (unsigned long)BLANK_CHECK_BUFFER_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  address = checkStart;
  remaining = checkLimit - checkStart;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hDevice, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_BLOCK_QUERY_FAILURE;
      goto done;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > BLANK_CHECK_BUFFER_SIZE) {
      chunk = BLANK_CHECK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hDevice, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_READ_FLASH_ERROR;
      goto done;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != 0xffU) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n"),
            (unsigned long long)(address + i), 0xffU, pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    remaining -= chunk;
  }

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }

  *pnError = nError;
  return ret;
}

static int
verify(
  ADMXRC3_HANDLE hDevice,
  ADMXRC3_BITSTREAM* pBitstream,
  unsigned int flashIndex,
  uint64_t verifyStart,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, chunk, nError = 0;
  size_t i, offset, remaining;
  int ret = EXIT_OK;
  
  pBuffer = (uint8_t*)malloc(VERIFY_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lx bytes\n"),
      (unsigned long)VERIFY_BUFFER_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  offset = 0;
  address = verifyStart;
  remaining = pBitstream->Length;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hDevice, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_BLOCK_QUERY_FAILURE;
      goto done;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > VERIFY_BUFFER_SIZE) {
      chunk = VERIFY_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hDevice, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_READ_FLASH_ERROR;
      goto done;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != pBitstream->Data[offset + i]) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n"),
            (unsigned long long)(address + i), pBitstream->Data[offset + i], pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    offset += (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
  }

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }

  *pnError = nError;
  return ret;
}

static int
readback(
  ADMXRC3_HANDLE hDevice,
  const ADMXRC3_FLASH_INFO* pFlashInfo,
  unsigned int flashIndex,
  const Region* pReadbackRegion,
  const TCHAR* pFilename)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, chunk;
  size_t n;
  FILE* pFile = NULL;
  int ret = EXIT_OK;

  if (pReadbackRegion->limit < pReadbackRegion->base) {
    _tprintf(_T("*** Readback region 0x%llX - 0x%llX wraps to address 0.\n"),
      (unsigned long long)pReadbackRegion->base, (unsigned long long)(pReadbackRegion->limit - 1));
    ret = EXIT_INVALID_RANGE;
    goto done;
  }
  if (0 == pReadbackRegion->size) {
    _tprintf(_T("*** Readback region has size of 0.\n"));
    ret = EXIT_INVALID_RANGE;
    goto done;
  }
  if ((pReadbackRegion->limit > pFlashInfo->UseableStart + pFlashInfo->UseableLength) ||
      (pReadbackRegion->base < pFlashInfo->UseableStart))
  {
    _tprintf(_T("*** Readback region 0x%llX - 0x%llX does not lie wholly within Flash useable area.\n"),
      (unsigned long long)pReadbackRegion->base, (unsigned long long)pReadbackRegion->limit - 1);
    ret = EXIT_INVALID_RANGE;
    goto done;
  }
  _tprintf(_T("Reading data from 0x%llX-0x%llX into '%s'...\n"),
    (unsigned long long)pReadbackRegion->base, (unsigned long long)(pReadbackRegion->limit - 1U), pFilename);

  pBuffer = (uint8_t*)malloc(READBACK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lX bytes\n"),
      (unsigned long)READBACK_BUFFER_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  if (_tfopen_s(&pFile, pFilename, _T("wb"))) {
    pFile = NULL;
    _tprintf(_T("*** Failed to open file '%s' for writing.\n"), pFilename);
    ret = EXIT_CREATE_FILE_ERROR;
    goto done;
  }

  address = pReadbackRegion->base;
  remaining = pReadbackRegion->size;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hDevice, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_BLOCK_QUERY_FAILURE;
      goto done;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > READBACK_BUFFER_SIZE) {
      chunk = READBACK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hDevice, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_READ_FLASH_ERROR;
      goto done;
    }

    n = fwrite(pBuffer, 1, (size_t)chunk, pFile); /* Cast is safe because we guarantee chunk < size of buffer */
    if (n != chunk) {
      _tprintf(_T("*** Failed to write all data to file, %lu(0x%lX) actual vs. %lu(0x%lX) expected.\n"),
        (unsigned long)n, (unsigned long)n, (unsigned long)chunk, (unsigned long)chunk);
      ret = EXIT_WRITE_FILE_ERROR;
      goto done;
    }

    address += chunk;
    remaining -= chunk;
  }

done:
  if (NULL != pFile) {
    fclose(pFile);
  }
  if (NULL != pBuffer) {
    free(pBuffer);
  }

  return ret;
}

/*
** Verify that boot flag lives entirely in region to be erased
*/
static int
checkBootFlagAddress(
  BootFlagType bootFlagType,
  const Region* bootFlagRegion,
  const Region* pEraseRegion)
{
  if (bootFlagType != BootFlagTypeNone) {
    if (bootFlagRegion->limit < bootFlagRegion->base) {
      _tprintf(_T("*** Boot flag @ 0x%llX - 0x%llX wraps to address 0.\n"),
        (unsigned long long)bootFlagRegion->base, (unsigned long long)(bootFlagRegion->limit - 1));
      return EXIT_SANITY_ERROR;
    }
    if ((bootFlagRegion->limit > pEraseRegion->limit) || (bootFlagRegion->base < pEraseRegion->base)) {
      _tprintf(_T("*** Boot flag @ 0x%llX - 0x%llX overruns range 0x%llX - 0x%llX.\n"),
        (unsigned long long)bootFlagRegion->base, (unsigned long long)(bootFlagRegion->limit - 1),
        (unsigned long long)pEraseRegion->base, (unsigned long long)(pEraseRegion->limit - 1));
      return EXIT_SANITY_ERROR;
    }
  }

  return 0;
}

/*
** Read .bit file into memory and strip off header.
** Verifies that FPGA identifier string in .bit file matches FPGA in device, unless bForce is set.
** Performs bit & word swapping if necessary.
*/
static int
readAndCheckBitstream(
  const TCHAR* pFilename,
  const ADMXRC3_FPGA_INFO* pFpgaInfo,
  bool_t bFpgaInfoValid,
  bool_t bForce,
  SwapFlags swapFlags,
  Region* pProgramRegion,
  const Region* pEraseRegion,
  ADMXRC3_BITSTREAM** ppBitstream)
{
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  ADMXRC3_STATUS status;
  uint32_t i;
  int ret = EXIT_OK;

  /* Load the .BIT file specified on the command line */
  status = ADMXRC3_LoadBitstream(pFilename, &pBitstream);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to load bitstream '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_READ_BIT_ERROR;
    goto done;
  }
  _tprintf(_T("Bitstream is for device %s, length %lu(0x%lX) bytes\n"),
    pBitstream->Identifier, (unsigned long)pBitstream->Length, (unsigned long)pBitstream->Length);

  pProgramRegion->size = pBitstream->Length;
  pProgramRegion->limit = pProgramRegion->base + pProgramRegion->size;
  if (pProgramRegion->limit < pProgramRegion->base) {
    _tprintf(_T("*** Bitstream @ 0x%llX - 0x%llX wraps to address 0.\n"),
      (unsigned long long)pProgramRegion->base, (unsigned long long)(pProgramRegion->limit - 1));
    ret = EXIT_SANITY_ERROR;
    goto done;
  }
  if (pProgramRegion->limit > pEraseRegion->limit) {
    _tprintf(_T("*** Bitstream @ 0x%llX - 0x%llX overruns range 0x%llX - 0x%llX.\n"),
        (unsigned long long)pProgramRegion->base, (unsigned long long)(pProgramRegion->limit - 1),
        (unsigned long long)pEraseRegion->base, (unsigned long long)(pEraseRegion->limit - 1));
    ret = EXIT_SANITY_ERROR;
    goto done;
  }

  if (bFpgaInfoValid) {
    if (!checkDeviceMatch(pBitstream->Identifier, pFpgaInfo->Identifier)) {
      if (!bForce) {
        _tprintf(_T("*** Bitstream device does not match target FPGA device - aborting\n"));
        ret = EXIT_FPGA_MISMATCH;
        goto done;
      } else {
        _tprintf(_T("+++ WARNING: Device information in bitstream does not match target FPGA device.\n"));
        _tprintf(_T("             Overriding due to +force option specified on command line.\n"));
      }
    }
  } else {
    assert(bForce);
    _tprintf(_T("+++ WARNING: Information for target FPGA was not found. Bitstream may or may not match actual target FPGA device.\n"));
    _tprintf(_T("             Overriding due to +force option specified on command line.\n"));
  }

  if (swapFlags & SwapFlagsBitSwap) {
    _tprintf(_T("Doing bit-swapping...\n"));
    for (i = 0; i < pBitstream->Length; i++) {
      uint8_t val8 = pBitstream->Data[i];
      pBitstream->Data[i] = (uint8_t)(((val8 & 0x80U) >> 7) | ((val8 & 0x40U) >> 5) | ((val8 & 0x20U) >> 3) | ((val8 & 0x10U) >> 1) |
                                      ((val8 & 0x08U) << 1) | ((val8 & 0x04U) << 3) | ((val8 & 0x02U) << 5) | ((val8 & 0x01U) << 7));
    }
  }
  switch (swapFlags & SwapFlagsByteSwapMask) {
  case 2U:
    _tprintf(_T("Doing 16-bit endian-swapping...\n"));
    for (i = 0; i < pBitstream->Length; i += 2) {
      uint8_t val8a = pBitstream->Data[i + 0];
      uint8_t val8b = pBitstream->Data[i + 1];
      pBitstream->Data[i + 0] = val8b;
      pBitstream->Data[i + 1] = val8a;
    }
    break;

  case 4U:
    _tprintf(_T("Doing 32-bit endian-swapping...\n"));
    for (i = 0; i < pBitstream->Length; i += 2) {
      uint8_t val8a = pBitstream->Data[i + 0];
      uint8_t val8b = pBitstream->Data[i + 1];
      uint8_t val8c = pBitstream->Data[i + 2];
      uint8_t val8d = pBitstream->Data[i + 3];
      pBitstream->Data[i + 0] = val8d;
      pBitstream->Data[i + 1] = val8c;
      pBitstream->Data[i + 2] = val8b;
      pBitstream->Data[i + 3] = val8a;
    }
    break;
  }

done:
  if (ret) {
    if (NULL != pBitstream) {
      ADMXRC3_UnloadBitstream(pBitstream);
    }
  } else {
    *ppBitstream = pBitstream;
  }
  return ret;
}

static int
programBootFlag(
  ADMXRC3_HANDLE hDevice,
  unsigned int flashIndex,
  BootFlagType bootFlagType,
  const Region* pBootFlagRegion,
  const ADMXRC3_BITSTREAM* pBitstream,
  bool_t* pbBootFlagVerify)
{
  ADMXRC3_STATUS status;
  uint32_t val32;
  uint8_t val8;
  int ret = EXIT_OK;

  /*
  ** For the 'program' command, the flag must be set in the Flash to cause the bitstream to be loaded
  ** at reset / power up.
  */
  switch (bootFlagType) {
  case BootFlagType24BitAddress:
  case BootFlagType30BitAddress:
    /*
    ** Set the flag that causes the FPGA to be loaded from flash,
    ** and bitstream length.
    */
    assert(pBootFlagRegion->size == 4);
    if (bootFlagType == BootFlagType24BitAddress) {
      val32 = (uint32_t)((pBitstream->Length - 1U) << 8);
    } else {
      val32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
    }
    _tprintf(_T("Setting boot flag at 0x%llX-0x%llX to 0x%08lX...\n"),
      (unsigned long long)pBootFlagRegion->base, (unsigned long long)(pBootFlagRegion->limit - 1U), (unsigned long)val32);
    val32 = cpu_to_le32(val32);
    status = ADMXRC3_WriteFlash(hDevice, flashIndex, 0, pBootFlagRegion->base, sizeof(val32), &val32);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      *pbBootFlagVerify = FALSE;
      ret = EXIT_WRITE_FLASH_ERROR;
      goto done;
    }
    break;

  case BootFlagType8Bit:
    /* Set the flag that causes the FPGA to be loaded from flash */
    assert(pBootFlagRegion->size == 1);
    val8 = 0x0U;
    _tprintf(_T("Setting boot flag at 0x%llX to 0x%02X...\n"), (unsigned long long)pBootFlagRegion->base, (unsigned int)val8);
    status = ADMXRC3_WriteFlash(hDevice, flashIndex, 0, pBootFlagRegion->base, sizeof(val8), &val8);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      *pbBootFlagVerify = FALSE;
      ret = EXIT_WRITE_FLASH_ERROR;
      goto done;
    }
    break;

  case BootFlagTypeNone:
    break;

  default:
    assert(FALSE);
    break;
  }

done:
  return ret;
}

static int
verifyBootFlag(
  ADMXRC3_HANDLE hDevice,
  unsigned int flashIndex,
  BootFlagType bootFlagType,
  const Region* pBootFlagRegion,
  ADMXRC3_BITSTREAM* pBitstream,
  bool_t* pbBootFlagOk)
{
  ADMXRC3_STATUS status;
  uint32_t expected32, bootFlag32;
  uint8_t expected8, bootFlag8;
  int ret = EXIT_OK;
  
  /*
  ** For the 'program' and 'verify' commands, the boot flag in the Flash that
  ** causes the bitstream to be loaded at reset / power up must be verified.
  */
  switch (bootFlagType) {
  case BootFlagType24BitAddress:
  case BootFlagType30BitAddress:
    /*
    ** Set the flag that causes the FPGA to be loaded from flash,
    ** and bitstream length.
    */
    assert(pBootFlagRegion->size == 4);
    if (bootFlagType == BootFlagType24BitAddress) {
      expected32 = (uint32_t)((pBitstream->Length - 1U) << 8);
    } else {
      expected32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
    }
    status = ADMXRC3_ReadFlash(hDevice, flashIndex, 0, pBootFlagRegion->base, sizeof(bootFlag32), &bootFlag32);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_READ_FLASH_ERROR;
    } else {
      bootFlag32 = le32_to_cpu(bootFlag32);
      if (bootFlag32 != expected32) {
        _tprintf(_T("*** Boot flag verification failed; expecting 0x%08lX, found 0x%08lX\n"),
          (unsigned long)expected32, (unsigned long)bootFlag32);
      } else {
        *pbBootFlagOk = TRUE;
      }
    }
    break;

  case BootFlagType8Bit:
    assert(pBootFlagRegion->size == 1);
    expected8 = 0x0U;
    status = ADMXRC3_ReadFlash(hDevice, flashIndex, 0, pBootFlagRegion->base, sizeof(bootFlag8), &bootFlag8);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back boot flag: %s\n"),
        ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_READ_FLASH_ERROR;
    } else {
      if (bootFlag8 != expected8) {
        _tprintf(_T("*** Boot flag verification failed; expecting 0x%02X, found 0x%02X\n"),
          (unsigned int)expected8, (unsigned int)bootFlag8);
      } else {
        *pbBootFlagOk = TRUE;
      }
    }
    break;

  case BootFlagTypeNone:
    assert(pBootFlagRegion->size == 0);
    *pbBootFlagOk = TRUE;
    break;

  default:
    assert(FALSE);
    break;
  }
  
  return ret;
}

static int
executeCommand(
  unsigned int index,
  bool_t bOpenByIndex,
  unsigned int serial,
  bool_t bOpenBySerial,
  CommandType command,
  unsigned int targetIndex,
  const TCHAR* pFilename,
  bool_t bFailsafeRegion,
  bool_t bForce,
  bool_t bNoVerify,
  unsigned int regionIndex,
  bool_t bRegionSpecified,
  const Region* pOverrideRegion, /* May be NULL */
  const Region* pReadbackRegion) /* May be NULL; only used by readback command */
{
  const double clockFreq = 33.0e6;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO	cardInfo;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  unsigned int flashIndex = (unsigned int)-1;
  Region eraseRegion = EMPTY_REGION;
  Region programRegion = EMPTY_REGION;
  Region bootFlagRegion = EMPTY_REGION;
  uint64_t nError = 0;
  bool_t bFpgaInfoValid = FALSE;
  BootFlagType bootFlagType = BootFlagTypeNone;
  bool_t bBootFlagVerify = TRUE;
  bool_t bBootFlagOk = FALSE;
  unsigned int numRegion, defaultRegionIndex, failsafeRegionIndex;
  SwapFlags swapFlags;
  int ret = EXIT_OK;

  if (bRegionSpecified && (command == CommandTypeInfo || command == CommandTypeReadback)) {
    _tprintf(_T("+++ WARNING: -region option has no effect on 'info' & 'readback' commands.\n"));
  }

  status = sampleOpenCardNoClearError(index, bOpenByIndex, serial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hDevice, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_CARDINFO_ERROR;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
  case ADMXRC3_MODEL_ADMVPX37V2:
    if (bFailsafeRegion) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = EXIT_FAILSAFE_NOT_SUPPORTED;
      goto done;
    }
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = EXIT_ILLEGAL_TARGET_INDEX;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADMXRC6TADV8:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = EXIT_ILLEGAL_TARGET_INDEX;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    if (bFailsafeRegion) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = EXIT_FAILSAFE_NOT_SUPPORTED;
      goto done;
    }
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = EXIT_ILLEGAL_TARGET_INDEX;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    if (targetIndex >= 2) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = EXIT_ILLEGAL_TARGET_INDEX;
      goto done;
    }
    flashIndex = targetIndex;
    break;

  default:
    _tprintf(_T("*** This utility must be run on one of the following cards:\n\n"));
    _tprintf(_T("\tADM-XRC-II\n"));
    _tprintf(_T("\tADPe-XRC-5T\n"));
    _tprintf(_T("\tADM-XRC-6TL\n"));
    _tprintf(_T("\tADM-XRC-6T1\n"));
    _tprintf(_T("\tADM-XRC-6T-DA1\n"));
    _tprintf(_T("\tADM-XRC-6TGE / ADM-XRC-6TGEL\n"));
    _tprintf(_T("\tADM-XRC-6T-ADV8\n"));
    _tprintf(_T("\tADPE-XRC-6T / ADPE-XRC-6T-L\n"));
    _tprintf(_T("\tADPE-XRC-6T-ADV (Controller)\n"));
    _tprintf(_T("\tADM-XRC-7K1\n"));
    _tprintf(_T("\tADM-XRC-7V1\n"));
    _tprintf(_T("\tADM-VPX3-7V2\n"));
    _tprintf(_T("\tADM-PCIE-7V3\n"));
    _tprintf(_T("\tADM-XRC-KU1\n"));
    _tprintf(_T("\tADM-PCIE-KU3\n"));
    _tprintf(_T("\tADM-PCIE-8V3\n"));
    _tprintf(_T("\tADM-PCIE-8K5\n"));
    ret = EXIT_UNSUPPORTED_MODEL;
    goto done;
  }

  status = ADMXRC3_GetFpgaInfo(hDevice, targetIndex, &fpgaInfo);
  if (status != ADMXRC3_SUCCESS) {
    if (!bForce) {
      _tprintf(_T("*** Failed to get FPGA %lu info: %s\n"), (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_FPGAINFO_ERROR;
      goto done;
    } else {
      _tprintf(_T("+++ WARNING: Failed to get FPGA %lu info: %s\n"), (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
      _tprintf(_T("             Overriding due to +force option specified on command line.\n"));
    }
  } else {
    _tprintf(_T("Target FPGA is device %s\n"), fpgaInfo.Identifier);
    bFpgaInfoValid = TRUE;
  }

  status = ADMXRC3_GetFlashInfo(hDevice, flashIndex, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_FLASHINFO_ERROR;
    goto done;
  }

  _tprintf(_T("Flash type is %s, %llu(0x%llX) kiB\n"), flashInfo.Identifier,
    (unsigned long long)(flashInfo.Size / 1024), (unsigned long long)(flashInfo.Size / 1024));
  _tprintf(_T("Useable region is 0x%llX-0x%llX\n"),
    (unsigned long long)flashInfo.UseableStart, (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1U));

  if (command == CommandTypeInfo) {
    goto done;
  }

  getModelDefaults(cardInfo.Model, &defaultRegionIndex, &failsafeRegionIndex, &numRegion, &bootFlagType, &swapFlags);
  if (!bRegionSpecified) {
    regionIndex = defaultRegionIndex;
  }
  if (bFailsafeRegion) {
    regionIndex = failsafeRegionIndex;
  }
  if (regionIndex >= numRegion) {
    _tprintf(_T("*** %u is not a valid Flash region for this model.\n    Region must be between 0 and %u (inclusive).\n"),
      regionIndex, numRegion - 1);
    ret = EXIT_ILLEGAL_REGION;
    goto done;
  }
  getEraseAddressRange(cardInfo.Model, &flashInfo, regionIndex, &eraseRegion);
  if (NULL != pOverrideRegion) {
    eraseRegion = *pOverrideRegion;
  }
  if (eraseRegion.limit > flashInfo.UseableStart + flashInfo.UseableLength ||
      eraseRegion.base < flashInfo.UseableStart)
  {
    _tprintf(_T("*** Selected range 0x%08llX - 0x%08llX overruns Flash useable area.\n"),
      (unsigned long long)eraseRegion.base, (unsigned long long)eraseRegion.limit - 1);
    ret = EXIT_INVALID_RANGE;
    goto done;
  }
  getProgramAddress(cardInfo.Model, eraseRegion.base, bootFlagType, &bootFlagRegion, &programRegion.base);

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    status = ADMXRC3_SetClockFrequency(hDevice, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_SET_CLOCK_FAILED;
      goto done;
    }
  }

  if (command == CommandTypeProgram || command == CommandTypeVerify) {
    /* Given that the user can override the Flash address range used for programming / verification,
     * we need to check that the boot flag is in range */
    ret = checkBootFlagAddress(
      bootFlagType,
      &bootFlagRegion,
      &eraseRegion);
    if (ret) {
      goto done;
    }

    /* We read the .bit file and check that it matches the FPGA in the device.
     * We also do bit and/or word-swapping of the SelectMap data if necessasry. */
    assert(NULL != pFilename);
    ret = readAndCheckBitstream(
      pFilename,
      &fpgaInfo,
      bFpgaInfoValid,
      bForce,
      swapFlags,
      &programRegion,
      &eraseRegion,
      &pBitstream);
    if (ret) {
      goto done;
    }
  }

  if (command == CommandTypeErase || command == CommandTypeProgram) {
    /* For the 'program' or 'erase' commands, a region of the Flash is first erased */
    _tprintf(_T("Erasing Flash from 0x%llX-0x%llX...\n"), (unsigned long long)eraseRegion.base, (unsigned long long)(eraseRegion.limit - 1U));
    ret = eraseFlash(hDevice, flashIndex, eraseRegion.base, eraseRegion.limit);
    if (ret) {
      goto done;
    }
  }

  if (command == CommandTypeProgram) {
    /* For the 'program' commands, the bitstream is written into a region of the Flash */
    _tprintf(_T("Writing bitstream at 0x%llX-0x%llX...\n"),
      (unsigned long long)programRegion.base, (unsigned long long)(programRegion.base + pBitstream->Length - 1U));
    ret = programBitstream(hDevice, pBitstream, flashIndex, programRegion.base);
    if (ret) {
      goto done;
    }
  }

  if ((command == CommandTypeErase && !bNoVerify) || command == CommandTypeChkblank) {
    /* For the 'erase' commands, a blank check is performed for a region of the Flash */
    _tprintf(_T("Blank checking from 0x%llX-0x%llX...\n"),
      (unsigned long long)eraseRegion.base, (unsigned long long)(eraseRegion.limit - 1U));
    ret = checkBlank(hDevice, flashIndex, eraseRegion.base, eraseRegion.limit, &nError);
    if (ret) {
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Blank check failed with %llu(0x%llX) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      ret = EXIT_NOT_BLANK;
      goto done;
    } else {
      _tprintf(_T("Blank check successful.\n"));
    }
  }

  if ((command == CommandTypeProgram && !bNoVerify) || command == CommandTypeVerify) {
    /* For the 'program' and 'verify' commands, the bitstream data is compared against the contents of a region of the Flash. */
    _tprintf(_T("Verifying from 0x%llX-0x%llX...\n"),
      (unsigned long long)programRegion.base, (unsigned long long)(programRegion.limit - 1U));
    ret = verify(hDevice, pBitstream, flashIndex, programRegion.base, &nError);
    if (ret) {
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Verification failed with %llu(0x%llX) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      ret = EXIT_VERIFY_FAILED;
      goto done;
    } else {
      _tprintf(_T("Verification successful.\n"));
    }
  }

  if (command == CommandTypeProgram) {
    /* For the 'program' command, we write the boot flag (if necessary, i.e. for models whose target FPGA configuration is performed by the Bridge). */
    ret = programBootFlag(
      hDevice,
      flashIndex,
      bootFlagType,
      &bootFlagRegion,
      pBitstream,
      &bBootFlagVerify);
  }

  if ((command == CommandTypeProgram && !bNoVerify) || command == CommandTypeVerify) {
    if (bBootFlagVerify) {
      ret = verifyBootFlag(
        hDevice,
        flashIndex,
        bootFlagType,
        &bootFlagRegion,
        pBitstream,
        &bBootFlagOk);
    }

    if (ret == EXIT_OK && bBootFlagOk) {
      _tprintf(_T("Boot flag OK. The bitstream will be loaded on reset / power up.\n"));
    } else {
      _tprintf(_T("*** The bitstream may not be loaded on reset / power up.\n"));
      if (ret == EXIT_OK) {
        ret = EXIT_BOOT_FLAG_VERIFY_ERROR;
      }
    }
  }

  if (command == CommandTypeReadback) {
    assert(NULL != pReadbackRegion);
    assert(NULL != pFilename);
    ret = readback(hDevice, &flashInfo, flashIndex, pReadbackRegion, pFilename);
    if (ret) {
      goto done;
    }
  }

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
  }

  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    if ((unsigned int)-1 != flashIndex) {
      status = ADMXRC3_SyncFlash(hDevice, flashIndex, 0);
      if (ADMXRC3_SUCCESS != status) {
        _tprintf(_T("+++ WARNING: Failed to synchronize Flash on exit: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      }
    }
    ADMXRC3_Close(hDevice);
  }

  return ret;
}

#if (defined(__VXWORKS__) || defined(__vxworks))
 
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL   (0x1U) /* Interpret 'index' as a serial number */
#define FLAG_FORCE     (0x10U) /* Do operation even if target FPGA device is wrong for specified .BIT file */
#define FLAG_FAILSAFE (0x100U) /* Program the failsafe image, rather than the default image */
#define FLAG_NOVERIFY (0x200U) /* For erase / program operations, do not perform blank check / verification at end */

int
admxrc3FlashHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  admxrc3FlashHelp          ()\n"));
  _tprintf(_T("  admxrc3FlashInfo          (indexOrSerial, flags, targetIndex)\n"));
  _tprintf(_T("  admxrc3FlashChkblankRegion(indexOrSerial, flags, targetIndex, regionIndex)\n"));
  _tprintf(_T("  admxrc3FlashChkblank      (indexOrSerial, flags, targetIndex[, regionStart, regionLength])\n"));
  _tprintf(_T("  admxrc3FlashEraseRegion   (indexOrSerial, flags, targetIndex, regionIndex)\n"));
  _tprintf(_T("  admxrc3FlashErase         (indexOrSerial, flags, targetIndex[, regionStart, regionLength])\n"));
  _tprintf(_T("  admxrc3FlashProgramRegion (indexOrSerial, flags, targetIndex, pBitFilename, regionIndex)\n"));
  _tprintf(_T("  admxrc3FlashProgram       (indexOrSerial, flags, targetIndex, pBitFilename[, regionStart, regionLength])\n"));
  _tprintf(_T("  admxrc3FlashVerifyRegion  (indexOrSerial, flags, targetIndex, pBitFilename, regionIndex)\n"));
  _tprintf(_T("  admxrc3FlashVerify        (indexOrSerial, flags, targetIndex, pBitFilename[, regionStart, regionLength])\n"));
  _tprintf(_T("  admxrc3FlashReadback      (indexOrSerial, flags, targetIndex, pOutputFilename, regionStart, regionLength)\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL 0x%X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_FORCE    0x%X => Do operation even if target FPGA device is wrong for specified .BIT file\n"), FLAG_FORCE);
  _tprintf(_T("  FLAG_FAILSAFE 0x%X => Program the failsafe image, rather than the default image\n"), FLAG_FAILSAFE);
  _tprintf(_T("  FLAG_NOVERIFY 0x%X => For erase / program operations, does not perform blank check / verification at the end\n"), FLAG_FAILSAFE);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

static void
checkFlags(
  unsigned int flags,
  bool_t* pbOpenBySerial,
  bool_t* pbForce,
  bool_t* pbFailsafeRegion,
  bool_t* pbNoVerify)
{
  if (flags & FLAG_BYSERIAL) {
    *pbOpenBySerial = TRUE;
  }

  if (flags & FLAG_FORCE) {
    *pbForce = TRUE;
  }

  if (flags & FLAG_FAILSAFE) {
    *pbFailsafeRegion = TRUE;
  }

  if (flags & FLAG_NOVERIFY) {
    *pbNoVerify = TRUE;
  }
}

/* Return information about Flash bank used by the target FPGA with the given index */
int
admxrc3FlashInfo(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex)
{
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bForce) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_FORCE (0x%X)."), FLAG_FORCE);
  }

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeInfo,
    targetIndex,
    NULL,
    FALSE,
    FALSE,
    FALSE,
    0,
    FALSE,
    NULL,
    NULL);
}

/* Blank-check a specific region */
int
admxrc3FlashChkblankRegion(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int regionIndex)
{
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashChkblankRegion: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeChkblank,
    targetIndex,
    NULL,
    bFailsafeRegion,
    bForce,
    FALSE,
    regionIndex,
    TRUE,
    NULL,
    NULL);
}

/* Blank-check default region, failsafe region or a specified address range */
int
admxrc3FlashChkblank(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int regionStart,  /* OPTIONAL */
  unsigned int regionLength) /* OPTIONAL If nonzero, specifies Flash address range to use */
{
  Region overrideRegion;
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  overrideRegion.base = regionStart;
  overrideRegion.size = regionLength;
  overrideRegion.limit = overrideRegion.base + overrideRegion.size;

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeChkblank,
    targetIndex,
    NULL,
    bFailsafeRegion,
    bForce,
    FALSE,
    0,
    FALSE,
    (overrideRegion.size == 0) ? NULL : &overrideRegion,
    NULL);
}

/* Erase a specific region */
int
admxrc3FlashEraseRegion(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int regionIndex)
{
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashEraseRegion: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeErase,
    targetIndex,
    NULL,
    bFailsafeRegion,
    bForce,
    bNoVerify,
    regionIndex,
    TRUE,
    NULL,
    NULL);
}

/* Erase default region, failsafe region or a specified address range */
int
admxrc3FlashErase(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int regionStart,  /* OPTIONAL */
  unsigned int regionLength) /* OPTIONAL If nonzero, specifies Flash address range to use */
{
  Region overrideRegion;
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  overrideRegion.base = regionStart;
  overrideRegion.size = regionLength;
  overrideRegion.limit = overrideRegion.base + overrideRegion.size;

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeErase,
    targetIndex,
    NULL,
    bFailsafeRegion,
    bForce,
    bNoVerify,
    0,
    FALSE,
    (overrideRegion.size == 0) ? NULL : &overrideRegion,
    NULL);
}

/* Program a specific region */
int
admxrc3FlashProgramRegion(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pBitFilename,
  unsigned int regionIndex)
{
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  if (NULL == pBitFilename) {
    _tprintf(_T("*** pBitFilename must not be NULL."));
    return EXIT_NULL_FILENAME;
  }

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashProgramRegion: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeProgram,
    targetIndex,
    pBitFilename,
    bFailsafeRegion,
    bForce,
    bNoVerify,
    regionIndex,
    TRUE,
    NULL,
    NULL);
}

/* Program default region, failsafe region or a specified address range */
int
admxrc3FlashProgram(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pBitFilename,
  unsigned int regionStart,  /* OPTIONAL */
  unsigned int regionLength) /* OPTIONAL If nonzero, specifies Flash address range to use */
{
  Region overrideRegion;
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  if (NULL == pBitFilename) {
    _tprintf(_T("*** admxrc3FlashProgram: pBitFilename must not be NULL."));
    return EXIT_NULL_FILENAME;
  }

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  overrideRegion.base = regionStart;
  overrideRegion.size = regionLength;
  overrideRegion.limit = overrideRegion.base + overrideRegion.size;

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeProgram,
    targetIndex,
    pBitFilename,
    bFailsafeRegion,
    bForce,
    bNoVerify,
    0,
    FALSE,
    (overrideRegion.size == 0) ? NULL : &overrideRegion,
    NULL);
}

/* Program a specific region */
int
admxrc3FlashVerifyRegion(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pBitFilename,
  unsigned int regionIndex)
{
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  if (NULL == pBitFilename) {
    _tprintf(_T("*** admxrc3FlashVerifyRegion: pBitFilename must not be NULL."));
    return EXIT_NULL_FILENAME;
  }

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashVerifyRegion: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeVerify,
    targetIndex,
    pBitFilename,
    bFailsafeRegion,
    bForce,
    FALSE,
    regionIndex,
    TRUE,
    NULL,
    NULL);
}

/* Verify default region, failsafe region or a specified address range */
int
admxrc3FlashVerify(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pBitFilename,
  unsigned int regionStart,  /* OPTIONAL */
  unsigned int regionLength) /* OPTIONAL If nonzero, specifies Flash address range to use */
{
  Region overrideRegion;
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  if (NULL == pBitFilename) {
    _tprintf(_T("*** admxrc3FlashVerify: pBitFilename must not be NULL."));
    return EXIT_NULL_FILENAME;
  }

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  overrideRegion.base = regionStart;
  overrideRegion.size = regionLength;
  overrideRegion.limit = overrideRegion.base + overrideRegion.size;

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeVerify,
    targetIndex,
    pBitFilename,
    bFailsafeRegion,
    bForce,
    FALSE,
    0,
    FALSE,
    (overrideRegion.size == 0) ? NULL : &overrideRegion,
    NULL);
}

/* Verify default region, failsafe region or a specified address range */
int
admxrc3FlashReadback(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pFilename,
  unsigned int regionStart,
  unsigned int regionLength)
{
  Region readbackRegion;
  bool_t bOpenBySerial = FALSE;
  bool_t bForce = FALSE;
  bool_t bFailsafeRegion = FALSE;
  bool_t bNoVerify = FALSE;

  if (NULL == pFilename) {
    _tprintf(_T("*** admxrc3FlashReadback(): pFilename must not be NULL."));
    return EXIT_NULL_FILENAME;
  }

  checkFlags(flags, &bOpenBySerial, &bForce, &bFailsafeRegion, &bNoVerify);

  if (bForce) {
    _tprintf(_T("+++ WARNING: admxrc3FlashReadback: Ignoring FLAG_FORCE (0x%X)."), FLAG_FORCE);
  }

  if (bFailsafeRegion) {
    _tprintf(_T("+++ WARNING: admxrc3FlashReadback: Ignoring FLAG_FAILSAFE (0x%X)."), FLAG_FAILSAFE);
  }

  if (bNoVerify) {
    _tprintf(_T("+++ WARNING: admxrc3FlashInfo: Ignoring FLAG_NOVERIFY (0x%X)."), FLAG_NOVERIFY);
  }

  readbackRegion.base = regionStart;
  readbackRegion.size = regionLength;
  readbackRegion.limit = readbackRegion.base + readbackRegion.size;

  return executeCommand(
    indexOrSerial,
    !bOpenBySerial,
    indexOrSerial,
    bOpenBySerial,
    CommandTypeReadback,
    targetIndex,
    pFilename,
    FALSE,
    FALSE,
    FALSE,
    0,
    FALSE,
    NULL,
    &readbackRegion);
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  TCHAR* pCommand;
  Arguments arguments = g_initArguments;
  Option options[7];
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  unsigned int index, serial, targetIndex = 0;
  CommandType command;
  Region overrideRegion = EMPTY_REGION;
  Region readbackRegion = EMPTY_REGION;
  bool_t bIndexSpecified, bSerialSpecified;
  bool_t bFailsafeSpecified, bFailsafeRegion;
  bool_t bForce, bNoVerify;
  bool_t bRegionSpecified, bRangeSpecified;
  TCHAR* pOverrideRange;
  TCHAR* pFilename = NULL;
  unsigned int regionIndex;
  int ret = EXIT_OK;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("failsafe");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("program failsafe region");
  options[3].key = _T("force");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("ignore device mismatch");
  options[4].key = _T("region");
  options[4].type = OptionInt;
  options[4].def.intVal = -1;
  options[4].meaning = _T("Flash region to use for chkblank/erase/program/verify");
  options[5].key = _T("range");
  options[5].type = OptionString;
  options[5].def.pStringVal = _T("");
  options[5].meaning = _T("Override address range to use for chkblank/erase/program/verify\n                                     Expects argument in the form of <address>,<length>");
  options[6].key = _T("noverify");
  options[6].type = OptionBoolean;
  options[6].def.booleanVal = FALSE;
  options[6].meaning = _T("Skip blank check/verification for erase/program commands");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.bNoStdOptions = TRUE;

  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("One of erase, info, program, readback, verify or chkblank");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("target");
  paramMeaning[1] = _T("Index of target FPGA");
  paramType[2] = OptionPath;
  paramInfo[2] = _T("file");
  paramMeaning[2] = _T("Bitstream filename (program or verify commands only)");
  paramType[3] = OptionUInt64;
  paramInfo[3] = _T("offset");
  paramMeaning[3] = _T("Offset for readback (readback command only)");
  paramType[4] = OptionUInt64;
  paramInfo[4] = _T("length");
  paramMeaning[4] = _T("Length for readback (readback command only)");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  if (sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }

  index = (unsigned int)options[0].value.uintVal;
  bIndexSpecified = options[0].specified;
  serial = (unsigned int)options[1].value.uintVal;
  bSerialSpecified = options[1].specified;
  bFailsafeRegion = options[2].value.booleanVal;
  bFailsafeSpecified = options[2].specified;
  bForce = options[3].value.booleanVal;
  regionIndex = options[4].value.intVal;
  bRegionSpecified = options[4].specified;
  bRangeSpecified = options[5].specified;
  pOverrideRange = options[5].value.pStringVal;
  bNoVerify = options[6].value.booleanVal;

  if (bRangeSpecified) {
    TCHAR* p;
    unsigned long long start, length;
    int n, m;

    p = _tcschr(pOverrideRange, _T(','));
    if (NULL == p) {
      _tprintf(_T("*** Malformed argument for '-range' option.\n"));
      ret = EXIT_BAD_RANGE;
      goto done;
    }
    n = _stscanf_s(pOverrideRange, _T("%llx"), &start);
    m = _stscanf_s(p + 1, _T("%llx"), &length);
    if (n != 1 || m != 1) {
      _tprintf(_T("*** Malformed argument for '-range' option.\n"));
      ret = EXIT_BAD_RANGE;
      goto done;
    }
    if (length == 0) {
      _tprintf(_T("*** Length part of argument for '-range' option cannot be 0.\n"));
      ret = EXIT_BAD_RANGE;
      goto done;
    }
    overrideRegion.base = (uint64_t)start;
    overrideRegion.limit = overrideRegion.base + (uint64_t)length;
    if (overrideRegion.limit < overrideRegion.base) {
      _tprintf(_T("*** Argument for '-range' option wraps to 0.\n"));
      ret = EXIT_BAD_RANGE;
      goto done;
    }
  }
  pCommand = arguments.param[0].pStringVal;

  if (bFailsafeRegion && bRegionSpecified) {
    _tprintf(_T("*** +failsafe option is mutually exclusive with -region option.\n"));
    ret = EXIT_ILLEGAL_FLAGS;
    goto done;
  }

  if (bFailsafeRegion && bRangeSpecified) {
    _tprintf(_T("*** +failsafe option is mutually exclusive with -range option.\n"));
    ret = EXIT_ILLEGAL_FLAGS;
    goto done;
  }

  if (bRegionSpecified && bRangeSpecified) {
    _tprintf(_T("*** -region option is mutually exclusive with -range option.\n"));
    ret = EXIT_ILLEGAL_FLAGS;
    goto done;
  }

  if (_tcsicmp(pCommand, _T("info")) == 0) {
    command = CommandTypeInfo;
  } else if (_tcsicmp(pCommand, _T("program")) == 0) {
    command = CommandTypeProgram;
  } else if (_tcsicmp(pCommand, _T("verify")) == 0) {
    command = CommandTypeVerify;
  } else if (_tcsicmp(pCommand, _T("erase")) == 0) {
    command = CommandTypeErase;
  } else if (_tcsicmp(pCommand, _T("chkblank")) == 0) {
    command = CommandTypeChkblank;
  } else if (_tcsicmp(pCommand, _T("readback")) == 0) {
    command = CommandTypeReadback;
  } else {
    _tprintf(_T("*** Please specify a valid command. Use flash /? to get help.\n"));
    ret = EXIT_UNRECOGNIZED_COMMAND;
    goto done;
  }

  if (command == CommandTypeProgram || command == CommandTypeVerify) {
    if (arguments.nParam < 3) {
      _tprintf(_T("*** program or verify command requires (1) target FPGA index and (2) .BIT filename.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    targetIndex = (unsigned int)arguments.param[1].intVal;
    pFilename = arguments.param[2].pPathVal;
  }

  if (command == CommandTypeInfo || command == CommandTypeErase || command == CommandTypeChkblank) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** chkblank, erase or info command requires target FPGA index.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    targetIndex = (unsigned int)arguments.param[1].intVal;
  }

  if (command == CommandTypeReadback) {
    if (arguments.nParam < 4) {
      _tprintf(_T("*** readback command requires (1) target FPGA index, (2) byte offset and (3) byte length.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    targetIndex = (unsigned int)arguments.param[1].intVal;
    pFilename = arguments.param[2].pPathVal;
    readbackRegion.base = arguments.param[3].uint64Val;
    readbackRegion.size = arguments.param[4].uint64Val;
    readbackRegion.limit = readbackRegion.base + readbackRegion.size;
  }

  if (bFailsafeSpecified && (command == CommandTypeInfo || command == CommandTypeReadback)) {
    _tprintf(_T("+++ WARNING: +failsafe & -failsafe options have no effect on 'info' & 'readback' commands.\n"));
  }

  if (bNoVerify && (command == CommandTypeInfo || command == CommandTypeChkblank || command == CommandTypeVerify || command == CommandTypeReadback)) {
    _tprintf(_T("+++ WARNING: +noverify option has no effect on 'chkblank', 'info', 'readback' & 'verify' commands.\n"));
  }

  ret = executeCommand(
    index,
    bIndexSpecified || !bSerialSpecified,
    serial,
    bSerialSpecified,
    command,
    targetIndex,
    pFilename,
    bFailsafeRegion,
    bForce,
    bNoVerify,
    regionIndex,
    bRegionSpecified,
    bRangeSpecified ? &overrideRegion : NULL,
    &readbackRegion);

done:
  sampleCleanupCommandLine(&arguments);

  return ret;
}

#endif
