/*
** iprog.c - utility to schedule or abort IPROG reconfiguration.
**
** (C) Copyright 2015-2016 Alpha Data
**
** Supported models:
**
**   o ADM-PCIE-7V3
**   o ADM-PCIE-KU3
**   o ADM-PCIE-8V3
**   o ADM-PCIE-8K5
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tcscpy_s(x, n, y) do { strncpy(x, y, n); (x)[(n) - 1] = '\0'; } while (0)
# define _tcslwr_s(x, n) strlwr_s(x, n)
# if defined(__VXWORKS__) || defined(__vxworks)
#   include <ctype.h>
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcsstr(x, y) strstr(x, y)
# define _tmain main
# define _tprintf printf
# define _stscanf_s sscanf
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>
#include <adb3.h>
#include <adb3/iprog.h>

#include <sdk_common.h>

#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_UNRECOGNIZED_COMMAND    (2) /* Invalid command passed as 1st positional argument */
#define EXIT_INSUFFICIENT_ARGS       (3) /* Not enough positional arguments after command */
#define EXIT_INVALID_WBSTAR          (4) /* Invalid WBSTAR address */
#define EXIT_DEVICE_OPEN_ERROR       (5) /* Failed to open ADMXRC3 device */
#define EXIT_ABORTIPROG_ERROR        (6) /* Call to ADB3_AbortIPROG failed */
#define EXIT_SCHEDULEIPROG_ERROR     (7) /* Call to ADB3_ScheduleIPROG failed */
#define EXIT_STATUSIPROG_ERROR       (8) /* Call to ADB3_StatusIPROG failed */
#define EXIT_UNSUPPORTED_MODEL       (9) /* IPROG operations not supported for this model */
#define EXIT_ILLEGAL_TARGET_INDEX   (10) /* IPROG operations not supported for the specified target FPGA index */
#define EXIT_ILLEGAL_REGION         (11) /* Flash region index is not valid for this model & target FPGA index */
#define EXIT_INVALID_FLASH_ADDRESS  (12) /* Flash address is not valid for this model & target FPGA index */

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

typedef enum {
  AddressTypeFlashAddr,
  AddressTypeRegion,
  AddressTypeWBSTAR
} AddressType;

#if !defined(_MSC_VER) && !(defined(__VXWORKS__) || defined(__vxworks))
static void
strlwr_s(
  TCHAR* pStr,
  size_t max)
{
  TCHAR c;

  while (max) {
    c = *pStr;
    if (c == _T('\0')) {
      break;
    }
    *pStr++ = tolower(c);
    max--;
  }
}
#endif

static int
flashAddressToWBSTAR(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex,
  uint64_t address,
  uint64_t* pWBSTAR)
{
  ADB3_UNIQUE_ID id;
  ADB3_STATUS status;
  
  status = ADB3_GetUniqueId(hDevice, &id);
  assert(status == ADB3_SUCCESS);
  
  switch (id.Model) {
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (0 != targetIndex) {
      _tprintf(_T("*** Target FPGA index must be 0 for this model.\n"));
      return EXIT_ILLEGAL_TARGET_INDEX;
    }
    if ((address & 0x1) || (address & 0xF8000000)) {
      _tprintf(_T("*** Flash address must be even and below 128 MiB.\n"));
      return EXIT_INVALID_FLASH_ADDRESS;
    }
    *pWBSTAR = ((address & 0x1FFFFFEU) >> 1) | /* Convert to 16-bit word address */
               (0x1U << 29) |                   /* Drive RS[1:0] */
               ((address & 0x6000000) << 5 );   /* Value for RS[1:0] */
    break;
    
  default:
    /* Not a supported model */
    _tprintf(_T("*** This model is not supported.\n"));
    return EXIT_UNSUPPORTED_MODEL;
  }
  
  return EXIT_OK;
}
    
static int
flashRegionToWBSTAR(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex,
  unsigned int regionIndex,
  uint64_t* pWBSTAR)
{
  ADB3_UNIQUE_ID id;
  ADB3_STATUS status;
  
  status = ADB3_GetUniqueId(hDevice, &id);
  assert(status == ADB3_SUCCESS);
  
  switch (id.Model) {
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
    if (0 != targetIndex) {
      _tprintf(_T("*** Target FPGA index must be 0 for this model.\n"));
      return EXIT_ILLEGAL_TARGET_INDEX;
    }
    if (regionIndex > 3) {
      _tprintf(_T("*** Flash region must 0, 1, 2 or 3.\n"));
      return EXIT_ILLEGAL_REGION;
    }
    *pWBSTAR = (0x1U << 29) |        /* Drive RS[1:0] */
               ((regionIndex & 0x3) << 30 /* Value for RS[1:0] = regionIndex */);
    break;
    
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (0 != targetIndex) {
      _tprintf(_T("*** Target FPGA index must be 0 for this model.\n"));
      return EXIT_ILLEGAL_TARGET_INDEX;
    }
    if (regionIndex > 1) {
      _tprintf(_T("*** Flash region must 0 or 1.\n"));
      return EXIT_ILLEGAL_REGION;
    }
    *pWBSTAR = (0x1U << 29) |        /* Drive RS[1:0] */
               ((regionIndex & 0x1) << 31 /* Value for RS[1] = regionIndex; RS[0] = 0 */);
    break;

  default:
    /* Not a supported model */
    _tprintf(_T("*** This model is not supported.\n"));
    return EXIT_UNSUPPORTED_MODEL;
  }
  
  return 0;
}
    
static int
validateWBSTAR(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex,
  uint64_t WBSTAR)
{
  ADB3_UNIQUE_ID id;
  ADB3_STATUS status;
  
  status = ADB3_GetUniqueId(hDevice, &id);
  assert(status == ADB3_SUCCESS);
  
  switch (id.Model) {
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (0 != targetIndex) {
      _tprintf(_T("*** Target FPGA index must be 0 for this model.\n"));
      return -9;
    }
    if (WBSTAR >= 0x100000000ULL) {
      _tprintf(_T("*** WBSTAR value must be 0xFFFFFFFF or less (32-bit unsigned integer).\n"));
      return -9;
    }
    break;
    
  default:
    /* Not a supported model */
    _tprintf(_T("*** This model is not supported.\n"));
    return -8;
  }
  
  return 0;
}

static int
abortIPROG(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex,
  bool_t bVerbose)
{
  ADB3_STATUS status;
  
  status = ADB3_AbortIPROG(hDevice, targetIndex);
  if (ADB3_SUCCESS != status) {
    _tprintf(_T("*** Failed to abort IPROG for target FPGA %u: %s\n"), targetIndex, ADB3_GetStatusString(status, TRUE));
    return EXIT_ABORTIPROG_ERROR;
  }
  if (bVerbose) {
    _tprintf(_T("Aborted IPROG.\n"));
  }
  
  return EXIT_OK;
}

static int
statusIPROG(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex)
{
  ADB3_STATUS status;
  uint32_t flags, delayMs;
  uint64_t address;
  uint32_t addrLo, addrHi;
  
  status = ADB3_StatusIPROG(hDevice, targetIndex, &flags, &address, &delayMs);
  if (ADB3_SUCCESS != status) {
    _tprintf(_T("*** Failed to get IPROG status for target FPGA %u: %s\n"), targetIndex, ADB3_GetStatusString(status, TRUE));
    return EXIT_STATUSIPROG_ERROR;
  }
  if (flags) {
    addrLo = (uint32_t)address;
    addrHi = (uint32_t)(address >> 32);
    assert(!((flags & ADB3_IPROG_FROMNOW) && (flags & ADB3_IPROG_ONSTOP)));
    if (flags & ADB3_IPROG_FROMNOW) {
      _tprintf(_T("IPROG is scheduled for target FPGA %u to reconfigure from WBSTAR address 0x%08lx_%08lx at %lu ms from now.\n"),
        targetIndex, (unsigned long)addrHi, (unsigned long)addrLo, (unsigned long)delayMs);
    } else if (flags & ADB3_IPROG_ONSTOP) {
      _tprintf(_T("IPROG is scheduled for target FPGA %u to reconfigure from WBSTAR address 0x%08lx_%08lx at %lu ms from device stop.\n"),
        targetIndex, (unsigned long)addrHi, (unsigned long)addrLo, (unsigned long)delayMs);
    } else {
      assert(FALSE);
    }
  } else {
    _tprintf(_T("IPROG is not scheduled for target FPGA %u.\n"), targetIndex);
  }
  
  return EXIT_OK;
}

static int
scheduleIPROG(
  ADB3_HANDLE hDevice,
  unsigned int targetIndex,
  bool_t bVerbose,
  uint32_t flags,
  AddressType addressType,
  uint64_t address,
  unsigned int regionIndex,
  uint32_t delayMs)
{
  ADB3_STATUS status;
  uint32_t addrLo, addrHi;
  int ret = EXIT_OK;
  
  switch (addressType) {
  case AddressTypeFlashAddr:
    /* Convert Flash byte address to a WBSTAR value. */
    ret = flashAddressToWBSTAR(hDevice, targetIndex, address, &address);
    if (ret) {
      return ret;
    }
    break;
    
  case AddressTypeRegion:
    /* Convert Flash region to a WBSTAR value. */
    ret = flashRegionToWBSTAR(hDevice, targetIndex, regionIndex, &address);
    if (ret) {
      return ret;
    }
    break;
    
  case AddressTypeWBSTAR:
    /* Check validity of WBSTAR value. */
    ret = validateWBSTAR(hDevice, targetIndex, address);
    if (ret) {
      return ret;
    }
    break;
    
  default:
    assert(FALSE);
    break;
  }
  
  addrLo = (uint32_t)address;
  addrHi = (uint32_t)(address >> 32);
  
  status = ADB3_ScheduleIPROG(hDevice, targetIndex, flags, address, (uint32_t)delayMs);
  if (ADB3_SUCCESS != status) {
    _tprintf(_T("*** Failed to schedule IPROG for target FPGA %u from %08lx_%08lx for %u ms from device stop: %s\n"),
      targetIndex, (unsigned long)addrHi, (unsigned long)addrLo, (unsigned int)delayMs, ADB3_GetStatusString(status, TRUE));
    return EXIT_SCHEDULEIPROG_ERROR;
  }
  if (bVerbose) {
    _tprintf(_T("Scheduled IPROG for target FPGA %u from %08lx_%08lx for %u ms from device stop.\n"),
      targetIndex, (unsigned long)addrHi, (unsigned long)addrLo, (unsigned int)delayMs);
  }
  
  return ret;
}

static ADB3_STATUS
sampleAdb3OpenCard(
  unsigned int index,
  bool_t bByIndex,
  unsigned int serial,
  bool_t bBySerial,
  bool_t bReadOnly,
  ADB3_HANDLE* phDevice)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice;
  ADB3_UNIQUE_ID id;
  unsigned int i;

  if (bByIndex && bBySerial) {
    /* Attempt to open card with specified index and serial number */
    status = ADB3_Open(index, bReadOnly, 0, &hDevice);
    if (status != ADB3_SUCCESS) {
      return status;
    }
    status = ADB3_GetUniqueId(hDevice, &id);
    if (status != ADB3_SUCCESS) {
      return status;
    }
    if (id.SerialNumber == serial) {
      *phDevice = hDevice;
      return ADB3_SUCCESS;
    }
    return ADB3_DEVICE_NOT_FOUND;
  }

  if (bBySerial) {
    /* Attempt to open card with specified serial number */
    for (i = 0; i < 256; i++) {
      status = ADB3_Open(i, bReadOnly, 0, &hDevice);
      if (status != ADB3_SUCCESS) {
        if (status == ADB3_DEVICE_NOT_FOUND) {
          return status;
        }
        continue;
      }
      status = ADB3_GetUniqueId(hDevice, &id);
      if (status != ADB3_SUCCESS) {
        continue;
      }
      if (id.SerialNumber == serial) {
        *phDevice = hDevice;
        return ADB3_SUCCESS;
      }
      ADB3_Close(hDevice);
    }
    return ADB3_DEVICE_NOT_FOUND;
  } else {
    /* Attempt to open card with specified index */
    return ADB3_Open(index, bReadOnly, 0, phDevice);
  }
}

#if (defined(__VXWORKS__) || defined(__vxworks))
 
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL (0x1U)  /* Interpret 'indexOrSerial' as a serial number */
#define FLAG_VERBOSE  (0x2U)  /* Verbose operation */

int
adb3IprogHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int adb3IprogHelp()\n"));
  _tprintf(_T("  int adb3IprogAbort(indexOrSerial, flags, targetIndex)\n"));
  _tprintf(_T("  int adb3IprogScheduleRegion(indexOrSerial, flags, targetIndex, regionIndex, delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogScheduleAddress(indexOrSerial, flags, targetIndex, flashAddress, delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogScheduleAddress(indexOrSerial, flags, targetIndex, flashAddress, delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogScheduleXAddress(indexOrSerial, flags, targetIndex, flashAddress(ULL), delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogScheduleWbstar(indexOrSerial, flags, targetIndex, wbstarValue, delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogScheduleXWbstar(indexOrSerial, flags, targetIndex, wbstarValue(ULL), delayMs, bScheduleFromNow)\n"));
  _tprintf(_T("  int adb3IprogStatus(indexOrSerial, flags, targetIndex)\n"));
  _tprintf(_T("'flashAddress(ULL)' & 'wbstarValue(ULL)' are unsigned long long parameters.\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL 0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_VERBOSE  0x%02X => Verbose operation\n"), FLAG_VERBOSE);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

static void
checkFlags(
  unsigned int flags,
  bool_t* pbOpenBySerial,
  bool_t* pbVerbose)
{
  if (flags & FLAG_BYSERIAL) {
    *pbOpenBySerial = TRUE;
  }

  if (flags & FLAG_VERBOSE) {
    *pbVerbose = TRUE;
  }
}

int
adb3IprogAbort(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bVerbose = FALSE;
  int ret = EXIT_OK; /* Assume success */
  
  checkFlags(flags, &bOpenBySerial, &bVerbose);
  
  status = sampleAdb3OpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = abortIPROG(hDevice, targetIndex, bVerbose);
  if (ret) {
    goto done;
  }

done:
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  return ret;
}

int
adb3IprogScheduleRegion(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int regionIndex,
  unsigned int delayMs,
  int bScheduleFromNow)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bVerbose = FALSE;
  int ret = EXIT_OK; /* Assume success */
  
  checkFlags(flags, &bOpenBySerial, &bVerbose);
  
  status = sampleAdb3OpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = scheduleIPROG(
    hDevice,
    targetIndex,
    bVerbose,
    bScheduleFromNow ? ADB3_IPROG_FROMNOW : ADB3_IPROG_ONSTOP,
    AddressTypeRegion,
    0,
    regionIndex,
    delayMs);
  if (ret) {
    goto done;
  }

done:
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  return ret;
}

int
adb3IprogScheduleXAddress(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned long long flashAddress,
  unsigned int delayMs,
  int bScheduleFromNow)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bVerbose = FALSE;
  int ret = EXIT_OK; /* Assume success */
  
  checkFlags(flags, &bOpenBySerial, &bVerbose);
  
  status = sampleAdb3OpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = scheduleIPROG(
    hDevice,
    targetIndex,
    bVerbose,
    bScheduleFromNow ? ADB3_IPROG_FROMNOW : ADB3_IPROG_ONSTOP,
    AddressTypeFlashAddr,
    (uint64_t)flashAddress,
    0,
    delayMs);
  if (ret) {
    goto done;
  }

done:
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  return ret;
}

int
adb3IprogScheduleAddress(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int flashAddress,
  unsigned int delayMs,
  int bScheduleFromNow)
{
  return adb3IprogScheduleRegion(indexOrSerial, flags, targetIndex, flashAddress, delayMs, bScheduleFromNow);
}

/* Entry point with WBSTAR value >= 64 bits, in case needed in future */
int
adb3IprogScheduleXWbstar(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned long long wbstarValueX,
  unsigned int delayMs,
  int bScheduleFromNow)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bVerbose = FALSE;
  int ret = EXIT_OK; /* Assume success */
  
  checkFlags(flags, &bOpenBySerial, &bVerbose);
  
  status = sampleAdb3OpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = scheduleIPROG(
    hDevice,
    targetIndex,
    bVerbose,
    bScheduleFromNow ? ADB3_IPROG_FROMNOW : ADB3_IPROG_ONSTOP,
    AddressTypeWBSTAR,
    (uint64_t)wbstarValueX,
    0,
    delayMs);
  if (ret) {
    goto done;
  }

done:
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  return ret;
}

int
adb3IprogScheduleWbstar(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  unsigned int wbstarValue,
  unsigned int delayMs,
  int bScheduleFromNow)
{
  return adb3IprogScheduleXWbstar(indexOrSerial, flags, targetIndex, wbstarValue, delayMs, bScheduleFromNow);
}

int
adb3IprogStatus(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex)
{
  ADB3_STATUS status;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bVerbose = FALSE;
  int ret = EXIT_OK; /* Assume success */
  
  checkFlags(flags, &bOpenBySerial, &bVerbose);
  
  status = sampleAdb3OpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = statusIPROG(hDevice, targetIndex);

done:
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  return ret;
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/

int _tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = EXIT_OK;
  ADB3_HANDLE hDevice = ADB3_HANDLE_INVALID_VALUE;
  ADB3_STATUS status;
  Arguments arguments = g_initArguments;
  Option options[3];
  OptionType paramType[4];
  TCHAR* paramInfo[4];
  TCHAR* paramMeaning[4];
  unsigned int serial = 0, index = 0;
  bool_t bVerbose;
  TCHAR* pCommand;
  TCHAR* pAddress;
  uint64_t address = 0;
  AddressType addressType = 0;
  unsigned int targetIndex, regionIndex = (unsigned int)-1;
  unsigned int delayMs = 0U;
  uint32_t flags = 0U;
  
  assert(sizeof(uint64_t) >= sizeof(unsigned long long));
  
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("verbose");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = 0;
  options[2].meaning = _T("TRUE => verbose output");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.bNoStdOptions = TRUE;
   
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("One of 'abort', 'from-now', 'on-stop', or 'status'");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("target");
  paramMeaning[1] = _T("Target FPGA index");
  paramType[2] = OptionString;
  paramInfo[2] = _T("address");
  paramMeaning[2] = _T("Address (hex) from which to reconfigure FPGA, e.g. address:6000000\n             or Flash region, e.g. region:3\n             or WBSTAR register value (hex), e.g. wbstar:E0000000");
  paramType[3] = OptionUInt;
  paramInfo[3] = _T("delay");
  paramMeaning[3] = _T("Delay in milliseconds");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 2;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  if (sampleParseCommandLine(_T("iprog"), argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }

  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bVerbose = options[2].value.booleanVal;

  pCommand = arguments.param[0].pStringVal;
  targetIndex = arguments.param[1].uintVal;
  delayMs = 30000U; /* Default delay */
  
  /* Validate command and parameters */
  if (_tcsicmp(pCommand, _T("abort")) == 0) {
    /* Nothing to check */
  } else if (_tcsicmp(pCommand, _T("from-now")) == 0 || _tcsicmp(pCommand, _T("on-stop")) == 0) {
    TCHAR tmp[32];
    unsigned long long addressUll;
    
    if (arguments.nParam < 3) {
      _tprintf(_T("*** Address/region/WBSTAR value required for 'from-now' or 'on-stop' command.\n"));
      _tprintf(_T("    Use 'iprog -h' for help.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    flags = (_tcsicmp(pCommand, _T("on-stop")) == 0) ? ADB3_IPROG_ONSTOP : ADB3_IPROG_FROMNOW;
    pAddress = arguments.param[2].pStringVal;
    _tcscpy_s(tmp, ARRAY_LENGTH(tmp), pAddress);
    _tcslwr_s(tmp, ARRAY_LENGTH(tmp));
    if (_tcsstr(tmp, _T("address:")) == tmp) {
      _stscanf_s(&tmp[8], _T("%llx"), &addressUll);
      address = (uint64_t)addressUll;
      addressType = AddressTypeFlashAddr;
    } else if (_tcsstr(tmp, _T("region:")) == tmp) {
      _stscanf_s(&tmp[7], _T("%u"), &regionIndex);
      addressType = AddressTypeRegion;
    } else if (_tcsstr(tmp, _T("wbstar:")) == tmp) {
      _stscanf_s(&tmp[7], _T("%llx"), &addressUll);
      address = (uint64_t)addressUll;
      addressType = AddressTypeWBSTAR;
    } else {
      _tprintf(_T("*** Invalid address for 'from-now' or 'on-stop' command.\n"));
      _tprintf(_T("    Use 'iprog -h' for help.\n"));
      ret = EXIT_INVALID_WBSTAR;
      goto done;
    }
    if (arguments.nParam >= 4) {
      delayMs = arguments.param[3].uintVal;
    }
  } else if (_tcsicmp(pCommand, _T("status")) == 0) {
    /* Nothing to check */
  } else {
    _tprintf(_T("*** Unrecognised command: %s\n"), pCommand);
    _tprintf(_T("    Use 'iprog -h' for help.\n"));
    ret = EXIT_UNRECOGNIZED_COMMAND;
    goto done;
  }
  
  status = sampleAdb3OpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  if (_tcsicmp(pCommand, _T("abort")) == 0) {
    ret = abortIPROG(hDevice, targetIndex, bVerbose);
    if (ret) {
      goto done;
    }
  } else if (_tcsicmp(pCommand, _T("on-stop")) == 0 || _tcsicmp(pCommand, _T("from-now")) == 0) {
    ret = scheduleIPROG(hDevice, bVerbose, targetIndex, flags, addressType, address, regionIndex, delayMs);
    if (ret) {
      goto done;
    }
  } else if (_tcsicmp(pCommand, _T("status")) == 0) {
    ret = statusIPROG(hDevice, targetIndex);
    if (ret) {
      goto done;
    }
  } else {
    assert(FALSE);
  }
  
done: 
  if (hDevice != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
