/*
** memtesth.c - Example that performs a host-driven (as opposed to FPGA-
**              driven) test of the memory devices connected to a target
**              FPGA.
**
** (C) Copyright 2010-2016 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define FALSE (0)
# define TRUE (1)

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

#define EXIT_OK                       (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE         (1) /* Illegal basic command-line syntax */
#define EXIT_DEVICE_OPEN_ERROR        (2) /* Failed to open ADMXRC3 device */
#define EXIT_CARDINFO_ERROR           (3) /* Failed to get information about ADMXRC3 device */
#define EXIT_ALLOCATION_FAILURE       (4) /* A memory allocation failed */
#define EXIT_SETCLOCKFREQUENCY_ERROR  (5) /* Failed to set clock generator frequency */
#define EXIT_CONFIGURATION_FAILED     (6) /* ADMXRC3_Configure{FromBuffer,FromFile} failed */
#define EXIT_BIT_FILENAME_FAILURE     (7) /* Failed to figure out full path of .bit file to use */
#define EXIT_MAPWINDOW_ERROR          (8) /* Failed to map window into process' virtual address space */
#define EXIT_LOCK_ERROR               (9) /* Failed to lock DMA data buffer using ADMXRC3_Lock */
#define EXIT_READDMA_ERROR           (10) /* Failed to read data from FPGA using ADMXRC3_ReadDMA(Ex) */
#define EXIT_WRITEDMA_ERROR          (11) /* Failed to write data to FPGA using ADMXRC3_WriteDMA(Ex) */
#define EXIT_VERIFICATION_FAILED     (12) /* Data read from FPGA contained at least one incorrect byte */
#define EXIT_TRAINING_FAILURE        (13) /* One or more memory banks failed to train */

#if !defined(USE_POINTER_ACCESS)
# define USE_POINTER_ACCESS (1)
#endif

/* Default maximum number of individual verification errors to report */
#define MAX_ERROR_REPORTED (20)

/* Maximum number of independent memory banks that this example supports */
#define MAX_BANKS (7)

/* Size of the direct slave window onto the FPGA registers */
#define DS_MAPPING_SIZE (0x400000U)

/* Chunk size used for transferring data during reads and writes of memory */
#define TRANSFER_CHUNK (0x100000U)

/* Arbirary nonzero starting value for 64-bit XOR LFSR that is used to generate 32-bit pseudorandom values */
const uint64_t g_lfsr64StartValue = (uint64_t)0xDEADBEEFCAFEFACEULL;

/* Attributes of direct slave window onto memory */
#define RAM_DS_WINDOW_BASE  0x200000U                  /* Base OCP address */
#define RAM_DS_WINDOW_ORDER 21U
#define RAM_DS_WINDOW_SIZE (1U << RAM_DS_WINDOW_ORDER) /* Size, in bytes*/
#define RAM_DS_WINDOW_MASK (RAM_DS_WINDOW_SIZE - 1U)

/* OCP addresses of FPGA registers */
#define REGADDR_RAM_DS_BANK          (0x300U)
#define REGADDR_RAM_DS_PAGE          (0x304U)
/* Per-bank registers (7 sets of 8 x 32-bit registers) */
#define REGADDR_BANK_CTRL(bank)      (0x320U + ((bank) << 5U))
#define REGADDR_BANK_OFFSET(bank)    (0x324U + ((bank) << 5U))
#define REGADDR_BANK_LENGTH(bank)    (0x328U + ((bank) << 5U))
#define REGADDR_BANK_INFO(bank)      (0x32CU + ((bank) << 5U))
#define REGADDR_BANK_STAT(bank)      (0x330U + ((bank) << 5U))
#define REGADDR_BANK_ERR_ADDR(bank)  (0x334U + ((bank) << 5U))
#define REGADDR_BANK_MUX_ERR(bank)   (0x338U + ((bank) << 5U))
#define REGADDR_BANK_DDR3_ERR(bank)  (0x33CU + ((bank) << 5U))

/* BANK_CTRL(x) fields */
#define REGBIT_BANK_CTRL_TEST        (0x1U << 8)  /* Write 1 to initiate FPGA-driven memory test */

/* BANK_INFO(x) fields */
#define REGMSK_BANK_INFO_WIDTH       (0xFFU << 16) /* Width of logical memory word in OCP */
#define REGPOS_BANK_INFO_WIDTH       (16)          /* Width of logical memory word in OCP */

/* BANK_STAT(x) fields */
#define REGBIT_BANK_STAT_WLEVELDONE  (0x1U << 0)  /* 1 => Write levelling done */
#define REGBIT_BANK_STAT_RLEVEL1DONE (0x1U << 1)  /* 1 => Read levelling done 1 */
#define REGBIT_BANK_STAT_RLEVEL2DONE (0x1U << 2)  /* 1 => Read levelling done 2 */
#define REGBIT_BANK_STAT_TRAINDONE   (0x1U << 3)  /* 1 => Everything finished training */
#define REGBIT_BANK_STAT_WLEVELERR   (0x1U << 8)  /* 1 => Error during write levelling */
#define REGBIT_BANK_STAT_RLEVEL1ERR  (0x1U << 9)  /* 1 => Error during read levelling 1 */
#define REGBIT_BANK_STAT_RLEVEL2ERR  (0x1U << 10) /* 1 => Error during read levelling 2 */
#define REGBIT_BANK_STAT_TRAINERR    (0x1U << 11) /* 1 => Error during training */
#define REGBIT_BANK_STAT_TESTDONE    (0x1U << 16) /* 1 => FPGA-driven memory test idle/completed */
#define REGBIT_BANK_STAT_TESTERR     (0x1U << 23) /* 1 => FPGA-driven memory test error */
#define REGMSK_BANK_STAT_ERRPHASE    (0xFU << 28) /* 1 => FPGA-driven memory test error phase */
#define REGPOS_BANK_STAT_ERRPHASE    (28)         /* 1 => FPGA-driven memory test error phase */

/* Returns the OCP address of a particular byte in a particular bank (applies only to DMA transfers) */
#define RAM_DMA_ADDRESS(bank, address) ((((uint64_t)(bank) + 1U) << 36U) + (address))

/* Information and state required by the {read|write}Bank functions */
typedef struct _TestState {
  ADMXRC3_HANDLE hDevice;
  unsigned int dmaChannel; /* DMA channel to use, when testing memory using DMA transfers */
  void* pFpgaSpace;
  unsigned int selectedBank; /* does not apply to DMA */
  uint32_t selectedPage; /* does not apply to DMA */
  admxrc3_time_t spinnerTime;
  unsigned int spinnerIndex;
} TestState;

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/* Writes a 32-bit value to an FPGA register; beware of write-buffering! */
static void
writeFpgaReg(
  TestState* pContext,
  size_t offset,
  uint32_t val32)
{
#if USE_POINTER_ACCESS
  volatile uint32_t* pFpgaSpace = (volatile uint32_t*)((uint8_t*)pContext->pFpgaSpace + offset);
  *pFpgaSpace = cpu_to_le32(val32);
#else
  val32 = cpu_to_le32(val32);
  ADMXRC3_Write(pContext->hDevice, 0, 0, offset, sizeof(uint32_t), &val32);
#endif
}

/* Reads a 32-bit value from an FPGA register */
static uint32_t
readFpgaReg(
  TestState* pContext,
  size_t offset)
{
  uint32_t val32;
#if USE_POINTER_ACCESS
  volatile uint32_t* pFpgaSpace = (volatile uint32_t*)((uint8_t*)pContext->pFpgaSpace + offset);
  val32 = *pFpgaSpace;
#else
  ADMXRC3_Read(pContext->hDevice, 0, 0, offset, sizeof(uint32_t), &val32);
#endif
  return le32_to_cpu(val32);
}

static void
spinner(
  unsigned int bank,
  bool_t bWrite,
  uint64_t address,
  TestState* pTestState)
{
  const TCHAR pattern[] = _T("|/-\\");
  admxrc3_time_t now;

  now = timeGet();
  if (timeAfter(now, pTestState->spinnerTime)) {
    pTestState->spinnerIndex++;
    if (pTestState->spinnerIndex >= ARRAY_LENGTH(pattern)) {
      pTestState->spinnerIndex = 0;
    }
    _tprintf(_T("%s bank %u @ 0x%llx %c                \r"), bWrite ? _T("Writing") : _T("Reading"), bank, (unsigned long long)address, pattern[pTestState->spinnerIndex]);
    fflush(stdout);
    pTestState->spinnerTime = timeAdd(pTestState->spinnerTime, timeMake(0, 100000, FALSE));
  }
}

/*
** Verify that a memory bank has been initialized and trained OK. Return
** value is:
**
**   -ve => error
**     0 => still training
**   +ve => training successful
*/
static int
checkBank(
  TestState* pContext,
  unsigned int bankIndex)
{
  uint32_t val32;

  val32 = readFpgaReg(pContext, REGADDR_BANK_STAT(bankIndex));
  if (val32 & REGBIT_BANK_STAT_TRAINDONE) {
    return 1;
  }
  if (val32 & REGBIT_BANK_STAT_TRAINERR) {
    return -1;
  }

  return 0;
}

/* Verify that all banks selected for test have been trained successfully */
static bool_t
checkBanks(
  TestState* pContext,
  unsigned int numBank,
  uint32_t bankMask)
{
  const uint32_t timeoutUs = 2000000U; /* 2s to complete training */
  admxrc3_time_t finish;
  uint32_t completedMask = 0, errorMask = 0;
  bool_t bOk = TRUE, bAllTrained;
  unsigned int i, bank;
  int res;

  finish = timeAdd(timeGet(), timeMake(timeoutUs / 1000000U, timeoutUs % 1000000U, TRUE));
  while (bOk) {
    bAllTrained = TRUE;
    completedMask = 0;
    errorMask = 0;
    for (bank = 0; bank < numBank; bank++) {
      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      res = checkBank(pContext, bank);
      if (res != 0) {
        completedMask |= 0x1U << bank;
        if (res < 0) {
          errorMask = 0x1U << bank;
        }
      } else {
        bAllTrained = FALSE;
      }
    }
    if (errorMask && bOk) {
      _tprintf(_T("*** Error during memory bank training\n"));
      bOk = FALSE;
      continue;
    }
    if (bAllTrained) {
      break;
    }
    if (timeAfter(timeGet(), finish)) {
      /* Timeout */
      _tprintf(_T("*** Timeout waiting for all banks to complete training\n"));
      bOk = FALSE;
      continue;
    }
    sleepMs(50);
  }
  if (!bOk) {
    _tprintf(_T("*** Completed bank mask is 0x%lx, error mask is 0x%lx\n"),
      (unsigned long)completedMask, (unsigned long)errorMask);
    for (i = 0; i < numBank; i++) {
      if (errorMask & (0x1U << i)) {
        _tprintf(_T("      BANK_STAT[%u]=0x%08lx\n"), i, (unsigned long)readFpgaReg(pContext, REGADDR_BANK_STAT(i)));
      }
    }
  }
  return bOk;
}

/* Select which bank of memory is being accessed by direct slave; does not apply to DMA transfers */
static void
selectBank(
  TestState* pContext,
  unsigned int bankIndex)
{
  writeFpgaReg(pContext, REGADDR_RAM_DS_BANK, bankIndex);
  readFpgaReg(pContext, REGADDR_RAM_DS_BANK);
}

/* Select which page of memory (in the selected bank) is being accessed by direct slave; does not apply to DMA transfers */
static void
selectPage(
  TestState* pContext,
  unsigned int pageIndex)
{
  writeFpgaReg(pContext, REGADDR_RAM_DS_PAGE, pageIndex);
  readFpgaReg(pContext, REGADDR_RAM_DS_PAGE);
}

/* Fill a buffer with a constant byte value */
static void
generateConstantByte(
  uint8_t* pBuffer,
  uint8_t val8,
  size_t n)
{
  while (n) {
    *pBuffer++ = val8;
    n--;
  }
}

/* Fill a buffer with own 32-bit address values */
static void
generateOwnAddress32(
  uint8_t* pBuffer,
  uint64_t address,
  size_t n)
{
  uint32_t val32 = (uint32_t)(address >> 2U);

  while (n) {
    *pBuffer++ = (uint8_t)(val32 >>  0U);
    *pBuffer++ = (uint8_t)(val32 >>  8U);
    *pBuffer++ = (uint8_t)(val32 >> 16U);
    *pBuffer++ = (uint8_t)(val32 >> 24U);
    n -= 4;
    val32++;
  }
}

/* Fill a buffer with own 32-bit values generated from a 64-bit LFSR */
static void
generateLfsr64(
  uint8_t* pBuffer,
  uint64_t* pLfsr64Value,
  size_t n)
{
  uint64_t lfsr64Val = *pLfsr64Value;
  uint8_t bit;

  while (n) {
    *pBuffer++ = (uint8_t)(lfsr64Val >>  0U);
    *pBuffer++ = (uint8_t)(lfsr64Val >>  8U);
    *pBuffer++ = (uint8_t)(lfsr64Val >> 16U);
    *pBuffer++ = (uint8_t)(lfsr64Val >> 24U);
    bit = (uint8_t)(((lfsr64Val >> 0U) ^ (lfsr64Val >> 1U) ^ (lfsr64Val >> 3U) ^ (lfsr64Val >> 4U)) & 0x1U);
    lfsr64Val = (lfsr64Val >> 1U) | ((uint64_t)bit << 63U);
    n -= 4;
  }
  *pLfsr64Value = lfsr64Val;
}

/* Reads a block of data from a memory bank, using direct slave or DMA */
static ADMXRC3_STATUS
readBank(
  TestState* pTestState,
  bool_t bUseDma,
  unsigned int bankIndex,
  uint64_t address,
  size_t n,
  void* pBuffer,
  ADMXRC3_BUFFER_HANDLE hBuffer)
{
  ADMXRC3_STATUS status;
  uint32_t page, offset;
  size_t chunk;
  uint8_t* p;

  if (bUseDma) {
    status = ADMXRC3_ReadDMALockedEx(pTestState->hDevice, pTestState->dmaChannel, 0, hBuffer, 0, n, RAM_DMA_ADDRESS(bankIndex, address));
    if (ADMXRC3_SUCCESS != status) {
      return status;
    }
  } else {
    if (pTestState->selectedBank != bankIndex) {
      selectBank(pTestState, bankIndex);
      pTestState->selectedBank = bankIndex;
    }
    page = (uint32_t)(address >> RAM_DS_WINDOW_ORDER);
    if (pTestState->selectedPage != page) {
      selectPage(pTestState, page);
      pTestState->selectedPage = page;
    }
    offset = (uint32_t)(address & RAM_DS_WINDOW_MASK);
    p = (uint8_t*)pBuffer;
    while (n) {
      chunk = RAM_DS_WINDOW_SIZE - offset;
      if (chunk > n) {
        chunk = n;
      }
#if USE_POINTER_ACCESS
      memcpy(p, (uint8_t*)pTestState->pFpgaSpace + offset + RAM_DS_WINDOW_BASE, chunk);
#else
      ADMXRC3_Read(pTestState->hDevice, 0, 0, offset + RAM_DS_WINDOW_BASE, chunk, p);
#endif
      address += chunk;
      n -= chunk;
      p += chunk;
      page++;
      selectPage(pTestState, page);
      pTestState->selectedPage = page;
      offset = 0;
    }
  }

  return ADMXRC3_SUCCESS;
}

/* Reads a block of data from a memory bank, using direct slave or DMA */
static ADMXRC3_STATUS
writeBank(
  TestState* pTestState,
  bool_t bUseDma,
  unsigned int bankIndex,
  uint64_t address,
  size_t n,
  const void* pBuffer,
  ADMXRC3_BUFFER_HANDLE hBuffer)
{
  ADMXRC3_STATUS status;
  uint32_t page, offset;
#if !USE_POINTER_ACCESS
  uint32_t dummy32;
#endif
  size_t chunk;
  uint8_t* p;

  if (bUseDma) {
    status = ADMXRC3_WriteDMALockedEx(pTestState->hDevice, pTestState->dmaChannel, 0, hBuffer, 0, n, RAM_DMA_ADDRESS(bankIndex, address));
    if (ADMXRC3_SUCCESS != status) {
      return status;
    }
  } else {
    if (pTestState->selectedBank != bankIndex) {
      selectBank(pTestState, bankIndex);
      pTestState->selectedBank = bankIndex;
    }
    page = (uint32_t)(address >> RAM_DS_WINDOW_ORDER);
    if (pTestState->selectedPage != page) {
      selectPage(pTestState, page);
      pTestState->selectedPage = page;
    }
    offset = (uint32_t)(address & RAM_DS_WINDOW_MASK);
    p = (uint8_t*)pBuffer;
    while (n) {
      chunk = RAM_DS_WINDOW_SIZE - offset;
      if (chunk > n) {
        chunk = n;
      }
#if USE_POINTER_ACCESS
      memcpy((uint8_t*)pTestState->pFpgaSpace + offset + RAM_DS_WINDOW_BASE, p, chunk);
	  (void)readFpgaReg(pTestState, RAM_DS_WINDOW_BASE);
#else
      ADMXRC3_Write(pTestState->hDevice, 0, 0, offset + RAM_DS_WINDOW_BASE, chunk, p);
      ADMXRC3_Read(pTestState->hDevice, 0, 0, RAM_DS_WINDOW_BASE, sizeof(uint32_t), &dummy32);
#endif
      address += chunk;
      n -= chunk;
      p += chunk;
      page++;
      selectPage(pTestState, page);
      pTestState->selectedPage = page;
      offset = 0;
    }
    readFpgaReg(pTestState, 0); /* Do a read of any FPGA register, to force all writes to be fully propagated */
  }

  return ADMXRC3_SUCCESS;
}

static void
verifyData(
  unsigned int bank,
  uint64_t address,
  const uint8_t* pReferenceBuffer,
  const uint8_t* pDataBuffer,
  size_t n,
  uint64_t maxError,
  uint64_t* pNumError)
{
  uint64_t numError = *pNumError;
  uint8_t expected, actual;

  while (n) {
    expected = *pReferenceBuffer++;
    actual = *pDataBuffer++;
    if (expected != actual) {
      if (numError < maxError) {
        _tprintf(_T("*** Verification error bank=%u address=0x%llx expected=0x%02x actual=0x%02x\n"),
          bank, (unsigned long long)address, (unsigned int)expected, (unsigned int)actual);
      } else if (numError == maxError) {
        _tprintf(_T("*** (More errors)\n"));
      }
      numError++;
    }
    address++;
    n--;
  }
  *pNumError = numError;
}


int
doMemTestH(
  ADMXRC3_HANDLE hDevice,
  const TCHAR* pBitFilename,
  uint32_t bankMask,
  bool_t bUseDma,
  uint64_t maxError,
  unsigned int numRep)
{
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_BANK_INFO bankInfo[MAX_BANKS];
  uint64_t bankSize[MAX_BANKS];
  uint32_t* pFpgaSpace = NULL;
  uint8_t* pCheckBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hCheckBuffer = 0;
  uint8_t* pGenerateBuffer = NULL;
  ADMXRC3_BUFFER_HANDLE hGenerateBuffer = 0;
  TestState testState;
  unsigned int bank, phase, direction, rep;
  uint64_t numError = 0;
  int ret = EXIT_OK;
  
  status = ADMXRC3_GetCardInfo(hDevice, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }

#if USE_POINTER_ACCESS
  /* Get the address of FPGA space */
  status = ADMXRC3_MapWindow(hDevice, 0, 0x0U, DS_MAPPING_SIZE, (void*)&pFpgaSpace);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to map FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_MAPWINDOW_ERROR;
    goto done;
  }
#endif

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hDevice, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1lf MHz: %s\n"), 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_SETCLOCKFREQUENCY_ERROR;
      goto done;
    }
  }

  /* Cannot test banks that do not exist */
  bankMask &= (uint32_t)(1U << cardInfo.NumMemoryBank) - 1U;
  /* Don't test banks that are not populated */
  bankMask &= cardInfo.MemoryBankPresent;
  for (bank = 0; bank < cardInfo.NumMemoryBank && bank < MAX_BANKS; bank++) {
    const TCHAR* pTypeStr = NULL;
    uint32_t type;

    status = ADMXRC3_GetBankInfo(hDevice, bank, &bankInfo[bank]);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("+++ Warning: failed to get memory bank %u information: %s\n"), bank, ADMXRC3_GetStatusString(status, TRUE));
      /* Don't test this bank, because we couldn't get any information about it */
      bankMask &= ~(1U << bank);
      continue;
    }
    /* Calculate bank size, in bytes */
    bankSize[bank] = bankInfo[bank].PhysicalSize * bankInfo[bank].PhysicalDataWidth / 8;
    type = bankInfo[bank].TypeMask;
    if (type & ADMXRC3_BANK_SDRAM_DDR3) {
      pTypeStr = _T("DDR-3 SDRAM");
    } else if (type & ADMXRC3_BANK_SSRAM_QDR) {
      pTypeStr = _T("QDR SSRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_DDR2) {
      pTypeStr = _T("DDR-2 SDRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_SDR) {
      pTypeStr = _T("SDR SDRAM");
    } else if (type & ADMXRC3_BANK_SSRAM_DDR2) {
      pTypeStr = _T("DDR-2 SSRAM");
    } else if (type & ADMXRC3_BANK_SDRAM_DDR) {
      pTypeStr = _T("DDR SDRAM");
    } else if (type & ADMXRC3_BANK_ZBT_P) {
      pTypeStr = _T("Pipelined ZBT");
    } else if (type & ADMXRC3_BANK_ZBT_FT) {
      pTypeStr = _T("Flowthrough ZBT");
    } else {
      pTypeStr = _T("Unknown type");
    }
    _tprintf(_T("Bank %02u: %s, %llu (0x%llx) kiB\n"), bank, pTypeStr, (unsigned long long)(bankSize[bank] >> 10U), (unsigned long long)(bankSize[bank] >> 10U));
  }
  _tprintf(_T("Bank test mask is 0x%04x\n"), (unsigned int)bankMask);

  pCheckBuffer = (uint8_t*)malloc(RAM_DS_WINDOW_SIZE);
  if (NULL == pCheckBuffer) {
    _tprintf(_T("*** Failed to allocate check buffer of 0x%x bytes.\n"), (unsigned int)RAM_DS_WINDOW_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  pGenerateBuffer = (uint8_t*)malloc(RAM_DS_WINDOW_SIZE);
  if (NULL == pGenerateBuffer) {
    _tprintf(_T("*** Failed to allocate generate buffer of 0x%x bytes.\n"), (unsigned int)RAM_DS_WINDOW_SIZE);
    ret = EXIT_ALLOCATION_FAILURE;
    goto done;
  }

  status = ADMXRC3_Lock(hDevice, pCheckBuffer, RAM_DS_WINDOW_SIZE, &hCheckBuffer);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to lock check buffer for DMA transfers: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_LOCK_ERROR;
    goto done;
  }

  status = ADMXRC3_Lock(hDevice, pGenerateBuffer, RAM_DS_WINDOW_SIZE, &hGenerateBuffer);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to lock generate buffer for DMA transfers: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_LOCK_ERROR;
    goto done;
  }

  if (NULL != pBitFilename) {
    status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pBitFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_CONFIGURATION_FAILED;
      goto done;
    }
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }

  testState.hDevice = hDevice;
  testState.pFpgaSpace = pFpgaSpace;
  testState.dmaChannel = 0;
  testState.selectedBank = 0;
  selectBank(&testState, testState.selectedBank);
  testState.selectedPage = 0;
  selectPage(&testState, testState.selectedPage);
  testState.spinnerTime = timeGet();
  testState.spinnerIndex = 0;

  /* Check the memory banks have been initialized and trained OK */
  if (!checkBanks(&testState, cardInfo.NumMemoryBank, bankMask)) {
    _tprintf(_T("*** FAILED - one or more banks not trained successfully\n"));
    ret = EXIT_TRAINING_FAILURE;
    goto done;
  }

  rep = 0;
  while (!numRep || rep < numRep) {
    bool_t bUseDmaRead = bUseDma;
    bool_t bUseDmaWrite = bUseDma;

    _tprintf(_T("Performing host-driven memory test (repetition %u)...\n"), rep);
    for (phase = 1; phase <= 2; phase++) {
      uint8_t byte = (uint8_t)((phase == 1) ? 0x55U : 0xAAU);
      _tprintf(_T("Phase %u - 0x%02X pattern               \n"), phase, byte);
      generateConstantByte(pGenerateBuffer, byte, RAM_DS_WINDOW_SIZE);
      for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
        uint64_t address, chunk;

        if (!(bankMask & (0x1U << bank))) {
          continue;
        }
        for (address = 0; address < bankSize[bank];) {
          spinner(bank, TRUE, address, &testState);
          chunk = bankSize[bank] - address;
          if (chunk > RAM_DS_WINDOW_SIZE) {
            chunk = RAM_DS_WINDOW_SIZE;
          }
          if (chunk > TRANSFER_CHUNK) {
            chunk = TRANSFER_CHUNK;
          }
          status = writeBank(&testState, bUseDmaWrite, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
            ret = EXIT_WRITEDMA_ERROR;
            goto done;
          }
          address += chunk;
        }
      }
      for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
        uint64_t address, chunk;

        if (!(bankMask & (0x1U << bank))) {
          continue;
        }
        for (address = 0; address < bankSize[bank];) {
          spinner(bank, FALSE, address, &testState);
          chunk = bankSize[bank] - address;
          if (chunk > RAM_DS_WINDOW_SIZE) {
            chunk = RAM_DS_WINDOW_SIZE;
          }
          if (chunk > TRANSFER_CHUNK) {
            chunk = TRANSFER_CHUNK;
          }
          status = readBank(&testState, bUseDmaRead, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
            ret = EXIT_READDMA_ERROR;
            goto done;
          }
          verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
          address += chunk;
        }
      }
      if (numError) {
        goto done;
      }
    }
    _tprintf(_T("Phase %u - own address pattern          \n"), phase);
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, TRUE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateOwnAddress32(pGenerateBuffer, address, (size_t)chunk);
        status = writeBank(&testState, bUseDmaWrite, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_WRITEDMA_ERROR;
          goto done;
        }
        address += chunk;
      }
    }
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, FALSE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateOwnAddress32(pGenerateBuffer, address, (size_t)chunk);
        status = readBank(&testState, bUseDmaRead, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_READDMA_ERROR;
          goto done;
        }
        verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
        address += chunk;
      }
    }
    phase++;
    if (numError) {
      goto done;
    }

    _tprintf(_T("Phase %u - pseudorandom data            \n"), phase);
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;
      uint64_t lfsr64Value = g_lfsr64StartValue;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, TRUE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateLfsr64(pGenerateBuffer, &lfsr64Value, (size_t)chunk);
        status = writeBank(&testState, bUseDmaWrite, bank, address, (size_t)chunk, pGenerateBuffer, hGenerateBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_WRITEDMA_ERROR;
          goto done;
        }
        address += chunk;
      }
    }
    for (bank = 0; bank < cardInfo.NumMemoryBank; bank++) {
      uint64_t address, chunk;
      uint64_t lfsr64Value = g_lfsr64StartValue;

      if (!(bankMask & (0x1U << bank))) {
        continue;
      }
      for (address = 0; address < bankSize[bank];) {
        spinner(bank, FALSE, address, &testState);
        chunk = bankSize[bank] - address;
        if (chunk > RAM_DS_WINDOW_SIZE) {
          chunk = RAM_DS_WINDOW_SIZE;
        }
        if (chunk > TRANSFER_CHUNK) {
          chunk = TRANSFER_CHUNK;
        }
        generateLfsr64(pGenerateBuffer, &lfsr64Value, (size_t)chunk);
        status = readBank(&testState, bUseDmaRead, bank, address, (size_t)chunk, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_READDMA_ERROR;
          goto done;
        }
        verifyData(bank, address, pGenerateBuffer, pCheckBuffer, (size_t)chunk, maxError, &numError);
        address += chunk;
      }
    }
    phase++;
    if (numError) {
      goto done;
    }
    rep++;
  }

  _tprintf(_T("Measuring throughput...                 \n"));
  for (direction = 0; direction < 2; direction++) {
    admxrc3_time_t start, finish, now;
    uint64_t total = 0;
    double elapsed, rate;

    start = timeGet();
    finish = timeAdd(start, timeMake(2, 0, FALSE));
    do {
      if (direction == 0) {
        status = writeBank(&testState, bUseDma, 0, 0, RAM_DS_WINDOW_SIZE, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to write memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_WRITEDMA_ERROR;
          goto done;
        }
      } else {
        status = readBank(&testState, bUseDma, 0, 0, RAM_DS_WINDOW_SIZE, pCheckBuffer, hCheckBuffer);
        if (ADMXRC3_SUCCESS != status) {
          _tprintf(_T("*** Failed to read memory: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
          ret = EXIT_READDMA_ERROR;
          goto done;
        }
      }
      total += RAM_DS_WINDOW_SIZE;
      now = timeGet();
    } while (timeBefore(now, finish));
    elapsed = timeToDouble(timeSubtract(now, start));
    rate = (double)total / elapsed;
    _tprintf(_T("Throughput %s is %.1f MiB/s\n"), (direction == 0) ? _T("from host to memory") : _T("from memory to host"), rate / 1048576.0);
  }

  _tprintf(_T("PASSED                                  \n"));

done:
  if (numError) {
    _tprintf(_T("*** FAILED with %llu (0x%llx) error(s)\n"), (unsigned long long)numError, (unsigned long long)numError);
    ret = -3;
  }

  if (0 != hGenerateBuffer) {
    ADMXRC3_Unlock(hDevice, hGenerateBuffer);
  }
  if (0 != hCheckBuffer) {
    ADMXRC3_Unlock(hDevice, hCheckBuffer);
  }
  if (NULL != pGenerateBuffer) {
    free(pGenerateBuffer);
  }
  if (NULL != pCheckBuffer) {
    free(pCheckBuffer);
  }
  if (NULL != pFpgaSpace) {
    status = ADMXRC3_UnmapWindow(hDevice, (void*)pFpgaSpace);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("+++ Failed to unmap FPGA space: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    }
  }
  
  return ret;
}

#if defined(__VXWORKS__) || defined(__vxworks)

/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/
 
/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL       (0x1U) /* Interpret 'index' as a serial number */
#define FLAG_NODMA          (0x10U) /* Use CPU-initiated I/O to transfer data (slow!) */

int
admxrc3MemTestHHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3MemTestHHelp()\n"));
  _tprintf(_T("  int admxrc3MemTestHX(indexOrSerial[, flags[, bankMask[, numRep[, maxError]]]])\n"));
  _tprintf(_T("  int admxrc3MemTestH(index[, bankMask[, bNoDma[, numRep[, maxError]]]])\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL       0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_NODMA          0x%02X => Use CPU-initiated I/O to transfer data (slow!)\n"), FLAG_NODMA);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3MemTestHX(
  unsigned int indexOrSerial, /* Index of device to open (OR serial number if 'flags' includes FLAG_BYSERIAL) */
  unsigned int flags,         /* See flags above */
  unsigned int bankMask,      /* Set of banks to test; 0 => all banks */
  int numRep,                 /* Number of repetitions to perform; < 0 => for ever, 0 or 1 => 1 repetition */
  unsigned int maxError)      /* Maximum errors to display in detail; 0 => default (20) */
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE;
  bool_t bUseDma = TRUE;
  TCHAR* pSdkBitDir = NULL;
  bool_t bShouldConfigure;
  TCHAR* pBitFilename = NULL;
  int ret = EXIT_OK;

  if (0 == bankMask) {
    bankMask = 0xFFFFFFFFU; /* Test all banks */
  }
  if (numRep < 0) {
    numRep = 0;
  } else if (0 == numRep) {
    numRep = 1;
  }
  if (0 == maxError) {
    maxError = MAX_ERROR_REPORTED;
  }
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial = TRUE;
  }
  if (flags & FLAG_NODMA) {
    bUseDma = FALSE;
  }
  
  pSdkBitDir = sampleGetSdkBitDir();
  
  status = sampleOpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index/serial %lu: %s\n"), (unsigned long)bOpenBySerial, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  pBitFilename = sampleConstructBitFilename(hDevice, 0, _T("uber"), _T("uber"), pSdkBitDir, &bShouldConfigure);
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doMemTestH(hDevice, bShouldConfigure ? pBitFilename : NULL, bankMask, bUseDma, maxError, (unsigned int)numRep);
  
done:
  if (NULL != pSdkBitDir) {
    free(pSdkBitDir);
  }
  if (NULL != pBitFilename) {
    free(pBitFilename);
  }
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

/* Entry point for compatibility with previous version */
int
admxrc3MemTestH(
  unsigned int index,    /* Index of device to open */
  unsigned int bankMask, /* Set of banks to test; 0 => all banks */
  int bNoDma,            /* TRUE => use CPU-initiated I/O to transfer data (slow!) */
  int numRep,            /* Number of repetitions to perform; -1 => for ever, 0 or 1 => 1 repetition */
  unsigned int maxError) /* Maximum errors to display in detail; 0 => default (20) */
{
  return admxrc3MemTestHX(index, bNoDma ? FLAG_NODMA : 0, bankMask, numRep, maxError);
}
#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR *argv[])
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  Option options[6];
  Arguments arguments = g_initArguments;
  const TCHAR* pBitFilename;
  bool_t bShouldConfigure;
  unsigned int index, serial, numRep;
  uint32_t bankMask;
  uint64_t maxError;
  bool_t bUseDma;
  int ret = EXIT_OK;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("banks");
  options[2].type = OptionUInt;
  options[2].def.uintVal = 0xFFFFU;
  options[2].meaning = _T("bitmask of banks to be tested");
  options[3].key = _T("dma");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = TRUE;
  options[3].meaning = _T("TRUE => use DMA to read and write memory");
  options[4].key = _T("maxerror");
  options[4].type = OptionUInt64;
  options[4].def.uint64Val = MAX_ERROR_REPORTED;
  options[4].meaning = _T("Maximum number of errors to display");
  options[5].key = _T("repeat");
  options[5].type = OptionUInt;
  options[5].def.uintVal = 1;
  options[5].meaning = _T("Number of repetitions to perform");
  arguments.nOption = 6;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  if (sampleParseCommandLine(_T("memtesth"), argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bankMask = options[2].value.uintVal;
  bUseDma = options[3].value.uintVal;
  maxError = options[4].value.uint64Val;
  numRep = options[5].value.uintVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }
  
  pBitFilename = sampleGetFpgaFilename(hDevice, 0, _T("uber"), _T("uber"), &arguments, &bShouldConfigure);
  if (NULL == pBitFilename) {
    _tprintf(_T("*** Failed to determine bitstream filename.\n"));
    ret = EXIT_BIT_FILENAME_FAILURE;
    goto done;
  }

  ret = doMemTestH(hDevice, bShouldConfigure ? pBitFilename : NULL, bankMask, bUseDma, maxError, numRep);
  
done:
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif