/*
** num_token.c - Common code for parsing character streams into numeric tokens
**
** (C) Copyright 2014 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stscanf_s(buffer, ...) _stscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istspace(x) isspace(x)
# define _stscanf_s sscanf
# define _totlower(c) tolower(c)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "sdk_common.h"
#include "num_token.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/* Pumps the state machine with a single character */
/* Returns TRUE if a token has just been fully parsed */
bool_t
sampleTokenPumpState(
  SampleTokenState* pState,
  TCHAR c,
  unsigned int flags)
{
  SampleTokenSubstate nextState = pState->substate;
  unsigned int nextStatus = pState->result.status;
  bool_t bParsed = FALSE, bStore = FALSE;

  if (c == _T('\0')) {
    c = _T(' ');
  } else {
    c = (TCHAR)_totlower(c);
  }
  switch (pState->substate) {
  case SampleTokenSubstateIdle:
    if (c == _T('\n') && (flags & SAMPLE_TOKEN_FLAG_NEWLINE_EMPTY_TOKEN)) {
      bParsed = TRUE;
      break;
    } else if (!_istspace(c)) {
      if (c >= _T('a') && c <= _T('f')) {
        nextState = SampleTokenSubstateHexadecimal;
        nextStatus = 0;
        bStore = TRUE;
      } else {
        if (c >= _T('0') && c <= _T('9')) {
          nextState = (flags & SAMPLE_TOKEN_FLAG_INTERPRET_HEX) ? SampleTokenSubstateHexadecimal : SampleTokenSubstateDecimal;
          nextStatus = 0;
          bStore = TRUE;
        } else {
          nextState = SampleTokenSubstateInvalid;
        }
      }
    }
    break;

  case SampleTokenSubstateDecimal:
    if (_totlower(c) == _T('x') && pState->position == 1 && pState->result.tokenBuf[0] == '0') {
      nextState = SampleTokenSubstateHexadecimal;
      pState->position = 0;
    } else if (c >= _T('a') && c <= _T('f')) {
      nextState = SampleTokenSubstateHexadecimal;
      bStore = TRUE;
    } else if (c >= _T('0') && c <= _T('9')) {
      nextState = SampleTokenSubstateDecimal;
      bStore = TRUE;
    } else if (_istspace(c)) {
      nextState = SampleTokenSubstateIdle;
      bParsed = TRUE;
    } else {
      nextState = SampleTokenSubstateInvalid;
      nextStatus |= SAMPLE_TOKEN_STATUS_MALFORMED;
    }
    break;

  case SampleTokenSubstateHexadecimal:
    if (_totlower(c) == _T('x') && pState->position == 1 && pState->result.tokenBuf[0] == '0') {
      nextState = SampleTokenSubstateHexadecimal;
      pState->position = 0;
    } else if ((c >= _T('a') && c <= _T('f')) || (c >= _T('0') && c <= _T('9'))) {
      nextState = SampleTokenSubstateHexadecimal;
      bStore = TRUE;
    } else if (c == _T('_')) {
      /* Ignore underscores inside hexadecimal tokens */
      nextState = SampleTokenSubstateHexadecimal;
    } else if (_istspace(c)) {
      nextState = SampleTokenSubstateIdle;
      bParsed = TRUE;
    } else {
      nextState = SampleTokenSubstateInvalid;
      nextStatus |= SAMPLE_TOKEN_STATUS_MALFORMED;
    }
    break;

  case SampleTokenSubstateInvalid:
    if (_istspace(c)) {
      nextState = SampleTokenSubstateIdle;
      bParsed = TRUE;
    } else {
      nextState = SampleTokenSubstateInvalid;
    }
    break;
  }

  if (bStore) {
    if (pState->position < ARRAY_LENGTH(pState->result.tokenBuf) - 1U) {
      pState->result.tokenBuf[pState->position] = c;
      pState->position++;
    } else {
      nextStatus |= SAMPLE_TOKEN_STATUS_TRUNCATED;
    }
  }

  if (bParsed) {
    pState->result.tokenBuf[pState->position] = _T('\0');
    pState->result.length = pState->position;
    pState->result.value = 0;
    if (pState->position > 0) {
      if (pState->substate == SampleTokenSubstateHexadecimal) {
        if (_stscanf_s(pState->result.tokenBuf, _T("%llX"), &pState->result.value) != 1) {
          nextStatus |= SAMPLE_TOKEN_STATUS_MALFORMED;
        }
      } else if (pState->substate == SampleTokenSubstateDecimal) {
        if (_stscanf_s(pState->result.tokenBuf, _T("%llu"), &pState->result.value) != 1) {
          nextStatus |= SAMPLE_TOKEN_STATUS_MALFORMED;
        }
      } else {
        pState->result.value = 0;
      }
      pState->position = 0;
    }
  }
  
  pState->substate = nextState;
  pState->result.status = nextStatus;

  return bParsed;
}

/* Pumps the state machine with a string */
/* Returns TRUE if a token has just been fully parsed, and if so, returns point at which parsing was complete in 'ppReached' */
bool_t
sampleTokenPumpString(
  SampleTokenState* pState,
  const TCHAR* pString,
  unsigned int flags,
  const TCHAR** ppReached)
{
  TCHAR c;
  
  do {
    c = *pString++;
    if (sampleTokenPumpState(pState, c, flags)) {
      if (NULL != ppReached) {
        *ppReached = pString;
      }
      return TRUE;
    }
  } while (c != _T('\0'));
  return FALSE;
}

/* Pumps the state machine with a file */
/* Returns TRUE if a token has just been fully parsed */
bool_t
sampleTokenPumpFile(
  SampleTokenState* pState,
  FILE* pFile,
  unsigned int flags)
{
  int c;

  do {
    c = fgetc(pFile);
    if (c == EOF) {
      if (sampleTokenPumpState(pState, _T('\0'), flags)) {
        return TRUE;
      }
    } else {
      if (sampleTokenPumpState(pState, (TCHAR)c, flags)) {
        return TRUE;
      }
    }
  } while (EOF != c);
  return FALSE;
}

extern bool_t
sampleTokenParsed(
  SampleTokenState* pState)
{
  return (pState->substate == SampleTokenSubstateIdle) && (pState->result.length > 0) ? TRUE : FALSE;
}
